"""
fs.base
=======

PyFilesystem base class


"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import abc
import os
import threading
import time
from functools import partial

from contextlib import closing
import itertools

import six

from . import copy
from . import errors
from . import iotools
from . import move
from . import tools
from . import walk
from . import wildcard
from .mode import validate_open_mode
from .path import abspath
from .path import join
from .path import normpath
from .time import datetime_to_epoch
from .walk import Walker


@six.add_metaclass(abc.ABCMeta)
class FS(object):
    """Base class for FS objects."""

    # This is the "standard" meta namespace.
    _meta = {}

    def __init__(self):
        self._closed = False
        self._lock = threading.RLock()
        super(FS, self).__init__()

    def __enter__(self):
        """Allow use of filesystem as a context manager."""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Close filesystem on exit."""
        self.close()

    @property
    def walk(self):
        """
        Get a :class:`~fs.walk.BoundWalker` object for this filesystem.

        """
        return Walker.bind(self)

    # ---------------------------------------------------------------- #
    # Required methods                                                 #
    # Filesystems must implement these methods.                        #
    # ---------------------------------------------------------------- #

    @abc.abstractmethod
    def getinfo(self, path, namespaces=None):
        """
        Get information regarding a resource (file or directory) on a
        filesystem.

        :param str path: A path to a resource on the filesystem.
        :param namespaces: Info namespaces to query (defaults to
            'basic').
        :type namespaces: list or None
        :returns: Resource information object.
        :rtype: :class:`~fs.info.Info`

        For more information regarding resource information see
        :ref:`info`.

        """

    @abc.abstractmethod
    def listdir(self, path):
        """
        Get a list of the resource names in a directory.

        :param str path: A path to a directory on the filesystem.
        :return: list of names, relative to ``path``.
        :rtype: list

        :raises fs.errors.DirectoryExpected: If `path` is not a
            directory.
        :raises fs.errors.ResourceNotFound: If `path` does not exist.

        This method will return a list of the resources in a directory.
        A 'resource' is a file, directory, or one of the other types
        defined in :class:`~fs.ResourceType`.

        """

    @abc.abstractmethod
    def makedir(self, path, permissions=None, recreate=False):
        """
        Make a directory, and return a :class:`~fs.subfs.SubFS` for
        the new directory.

        :param str path: Path to directory from root.
        :param permissions: :class:`~fs.permissions.Permissions`
            instance.
        :type permissions: Permissions
        :param bool recreate: Do not raise an error if the directory
            exists.
        :rtype: :class:`~fs.subfs.SubFS`

        :raises fs.errors.DirectoryExists: if the path already exists.
        :raises fs.errors.ResourceNotFound: if the path is not found.

        """

    @abc.abstractmethod
    def openbin(self, path, mode="r", buffering=-1, **options):
        """
        Open a binary file-like object.

        :param str path: A path on the filesystem.
        :param str mode: Mode to open file (must be a valid non-text
            mode). Since this method only opens binary files, the `b` in
            the mode string is implied.
        :param buffering: Buffering policy (-1  to use default
            buffering, 0 to disable buffering, or positive integer to
            indicate buffer size).
        :type buffering: int
        :param options: Keyword parameters for any additional
            information required by the filesystem (if any).
        :rtype: file object

        :raises fs.errors.FileExpected: If the path is not a file.
        :raises fs.errors.FileExists: If the file exists, and
            *exclusive mode* is specified (`x` in the mode).
        :raises fs.errors.ResourceNotFound: If `path` does not exist.

        """

    @abc.abstractmethod
    def remove(self, path):
        """
        Remove a file.

        :param str path: Path to the file you want to remove.

        :raises fs.errors.FileExpected: if the path is a directory.
        :raises fs.errors.ResourceNotFound: if the path does not
            exist.

        """

    @abc.abstractmethod
    def removedir(self, path):
        """
        Remove a directory from the filesystem.

        :param str path: Path of the directory to remove.

        :raises fs.errors.DirectoryNotEmpty: If the directory is not
            empty (see :meth:`~fs.base.removetree` if you want to
            remove the directory contents).
        :raises fs.errors.DirectoryExpected: If the path is not a
            directory.
        :raises fs.errors.ResourceNotFound: If the path does not
            exist.
        :raises fs.errors.RemoveRootError: If an attempt is made to
            remove the root directory (i.e. `'/'`).
        """

    @abc.abstractmethod
    def setinfo(self, path, info):
        """
        Set info on a resource.

        :param str path: Path to a resource on the filesystem.
        :param dict info: Dict of resource info.

        :raises fs.errors.ResourceNotFound: If `path` does not exist
            on the filesystem.

        This method is the compliment to :class:`~fs.base.getinfo` and
        is used to set info values on a resource.

        The ``info`` dict should be in the same format as the raw
        info returned by ``getinfo(file).raw``. Here's an example::

            details_info = {
                "details":
                {
                    "modified_time": time.time()
                }
            }
            my_fs.setinfo('file.txt', details_info)

        """

    # ---------------------------------------------------------------- #
    # Optional methods                                                 #
    # Filesystems *may* implement these methods.                       #
    # ---------------------------------------------------------------- #

    def appendbytes(self, path, data):
        """
        Append bytes to the end of a file. Creating the file if it
        doesn't already exists.

        :param str path: Path to a file.
        :param bytes data: Bytes to append.
        :raises TypeError: if ``data`` is not bytes.
        :raises fs.errors.ResourceNotFound: if a parent directory of
            ``path`` does not exist.

        """
        if not isinstance(data, bytes):
            raise TypeError('must be bytes')
        with self._lock:
            with self.open(path, 'ab') as append_file:
                append_file.write(data)

    def appendtext(self,
                   path,
                   text,
                   encoding='utf-8',
                   errors=None,
                   newline=''):
        """
        Append text to a file. Creating the file if it doesn't already
        exists.

        :param str path: Path to a file.
        :param str text: Text to append.
        :raises TypeError: if ``text`` is not bytes.
        :raises fs.errors.ResourceNotFound: if a parent directory of
            ``path`` does not exist.

        """
        if not isinstance(text, six.text_type):
            raise TypeError('must be unicode string')
        with self._lock:
            with self.open(path,
                           'at',
                           encoding=encoding,
                           errors=errors,
                           newline=newline) as append_file:
                append_file.write(text)

    def close(self):
        """
        Close the filesystem and release any resources.

        It is important to call this method when you have finished
        working with the filesystem. Some filesystems may not finalize
        changes until they are closed (archives for example). You may
        call this method explicitly (it is safe to call close multiple
        times), or you can use the filesystem as a context manager to
        automatically close.

        Here's an example of automatically closing a filesystem::

            with OSFS('~/Desktop') as desktop_fs:
                desktop_fs.settext(
                    'note.txt',
                    "Don't forget to tape Game of Thrones"
                )

        If you attempt to use a filesystem that has been closed, a
        :class:`~fs.errors.FilesystemClosed` exception will be thrown.

        """
        self._closed = True

    def copy(self, src_path, dst_path, overwrite=False):
        """
        Copy file contents from ``src_path`` to ``dst_path``.

        :param src_path: Path of source file.
        :type src_path: str
        :param dst_path: Path to destination file.
        :type dst_path: str
        :raises fs.errors.DestinationExists: If ``dst_path`` exists,
            and overwrite == ``False``.
        :raises fs.errors.ResourceNotFound: If a parent directory of
            ``dst_path`` does not exist.

        """
        with self._lock:
            if not overwrite and self.exists(dst_path):
                raise errors.DestinationExists(dst_path)
            with closing(self.open(src_path, 'rb')) as read_file:
                self.setbinfile(dst_path, read_file)

    def copydir(self, src_path, dst_path, create=False):
        """
        Copy the contents of ``src_path`` to ``dst_path``.

        :param str src_path: Source directory.
        :param str dst_path: Destination directory.
        :param bool create: If ``True`` then ``src_path`` will be
            created if it doesn't already exist.
        :raises fs.errors.ResourceNotFound: If the destination
            directory does not exist, and ``create`` is not True.

        """
        with self._lock:
            if not create and not self.exists(dst_path):
                raise errors.ResourceNotFound(dst_path)
            if not self.getinfo(src_path).is_dir:
                raise errors.DirectoryExpected(src_path)
            copy.copy_dir(
                self,
                src_path,
                self,
                dst_path
            )

    def create(self, path, wipe=False):
        """
        Create an empty file.

        :param str path: Path to new file in filesystem.
        :param bool wipe: Truncate any existing file to 0 bytes.
        :returns: ``True`` if file was created, ``False`` if it already
            existed.
        :rtype: bool

        The default behavior is to create a new file if one doesn't
        already exist. If ``wipe == True`` an existing file will be
        truncated.

        """
        with self._lock:
            if not wipe and self.exists(path):
                return False
            with closing(self.open(path, 'wb')):
                pass
            return True

    def desc(self, path):
        """
        Return a short descriptive text regarding a path.

        :param str path: A path to a resource on the filesystem.
        :rtype: str

        """
        if not self.exists(path):
            raise errors.ResourceNotFound(path)
        try:
            syspath = self.getsyspath(path)
        except (errors.ResourceNotFound, errors.NoSysPath):
            return "{} on {}".format(path, self)
        else:
            return syspath

    def exists(self, path):
        """
        Check if a path maps to a resource.

        :param str path: Path to a resource
        :rtype: bool

        A ``path`` exists if it maps to any resource (including
        a directory).

        """
        try:
            self.getinfo(path)
        except errors.ResourceNotFound:
            return False
        else:
            return True

    def filterdir(self,
                  path,
                  files=None,
                  dirs=None,
                  exclude_dirs=None,
                  exclude_files=None,
                  namespaces=None,
                  page=None):
        """
        Get an iterator of resource info, filtered by file patterns.

        :param str path: A path to a directory on the filesystem.
        :param list files: A list of unix shell-style patterns to filter
            file names, e.g. ``['*.py']``.
        :param list dirs: A list of unix shell-style wildcards to
            filter directory names.
        :param list exclude_dirs: An optional list of patterns used to
            exclude directories
        :param list exclude_files: An optional list of patterns used to
            exclude files.
        :param list namespaces: A list of namespaces to include in
            the resource information.
        :param page: May be a tuple of ``(<start>, <end>)`` indexes to
            return an iterator of a subset of the resource info, or
            ``None`` to iterate over the entire directory. Paging a
            directory scan may be necessary for very large directories.
        :type page: tuple or None
        :return: An iterator of :class:`~fs.info.Info` objects.
        :rtype: iterator

        This method enhances :meth:`~fs.base.FS.scandir` with additional
        filtering functionality.

        """

        resources = self.scandir(path, namespaces=namespaces)
        filters = []

        def match_dir(patterns, info):
            """Pattern match info.name"""
            return info.is_file or self.match(patterns, info.name)

        def match_file(patterns, info):
            """Pattern match info.name"""
            return info.is_dir or self.match(patterns, info.name)

        def exclude_dir(patterns, info):
            """Pattern match info.name"""
            return info.is_file or not self.match(patterns, info.name)

        def exclude_file(patterns, info):
            """Pattern match info.name"""
            return info.is_dir or not self.match(patterns, info.name)

        if files:
            filters.append(partial(match_file, files))
        if dirs:
            filters.append(partial(match_dir, dirs))
        if exclude_dirs:
            filters.append(partial(exclude_dir, exclude_dirs))
        if exclude_files:
            filters.append(partial(exclude_file, exclude_files))

        if filters:
            resources = (
                info
                for info in resources
                if all(_filter(info) for _filter in filters)
            )

        iter_info = iter(resources)
        if page is not None:
            start, end = page
            iter_info = itertools.islice(iter_info, start, end)
        return iter_info

    def getbytes(self, path):
        """
        Get the contents of a file as bytes.

        :param str path: A path to a readable file on the filesystem.
        :returns: file contents
        :rtype: bytes

        :raises fs.errors.ResourceNotFound: If ``path`` does not
            exist.

        """
        with closing(self.open(path, mode='rb')) as read_file:
            contents = read_file.read()
        return contents

    def gettext(self, path, encoding=None, errors=None, newline=''):
        """
        Get the contents of a file as a string.

        :param str path: A path to a readable file on the filesystem.
        :param str encoding: Encoding to use when reading contents in
            text mode.
        :param str errors: Unicode errors parameter.
        :param str newline: Newlines parameter.
        :returns: file contents.
        :raises fs.errors.ResourceNotFound: If ``path`` does not
            exist.

        """
        with closing(
            self.open(
                path,
                mode='rt',
                encoding=encoding,
                errors=errors,
                newline=newline)
        ) as read_file:
            contents = read_file.read()
        return contents

    def getmeta(self, namespace="standard"):
        """
        Get meta information regarding a filesystem.

        :param keys: A list of keys to retrieve, or None for all keys.
        :type keys: list or None
        :param str namespace: The meta namespace (default is
            `"standard"`).
        :rtype: dict

        Meta information is associated with a *namespace* which may be
        specified with the `namespace` parameter. The default namespace,
        ``"standard"``, contains common information regarding the
        filesystem's capabilities. Some filesystems may provide other
        namespaces which expose less common or implementation specific
        information. If a requested namespace is not supported by
        a filesystem, then an empty dictionary will be returned.

        The ``"standard"`` namespace supports the following keys:

        =================== ============================================
        key                 Description
        ------------------- --------------------------------------------
        case_insensitive    True if this filesystem is case insensitive.
        invalid_path_chars  A string containing the characters that may
                            may not be used on this filesystem.
        max_path_length     Maximum number of characters permitted in a
                            path, or None for no limit.
        max_sys_path_length Maximum number of characters permitted in
                            a sys path, or None for no limit.
        network             True if this filesystem requires a network.
        read_only           True if this filesystem is read only.
        supports_rename     True if this filesystem supports an
                            os.rename operation.
        =================== ============================================

        Most builtin filesystems will provide all these keys, and third-
        party filesystems should do so whenever possible, but a key may
        not be present if there is no way to know the value.

        .. note::
            Meta information is constant for the lifetime of the
            filesystem, and may be cached.


        """
        if namespace == 'standard':
            meta = self._meta.copy()
        else:
            meta = {}
        return meta

    def getsize(self, path):
        """
        Get the size (in bytes) of a resource.

        :param str path: A path to a resource.
        :rtype: int

        The *size* of a file is the total number of readable bytes,
        which may not reflect the exact number of bytes of reserved
        disk space (or other storage medium).

        The size of a directory is the number of bytes of overhead
        use to store the directory entry.

        """
        size = self.getdetails(path).size
        return size

    def getsyspath(self, path):
        """
        Get an *system path* to a resource.

        :param str path: A path on the filesystem.
        :rtype: str
        :raises fs.errors.NoSysPath: If there is no corresponding system path.

        A system path is one recognized by the OS, that may be used
        outside of PyFilesystem (in an application or a shell for
        example). This method will get the corresponding system path
        that would be referenced by ``path``.

        Not all filesystems have associated system paths. Network and
        memory based filesystems, for example, may not physically store
        data anywhere the OS knows about. It is also possible for some
        paths to have a system path, whereas others don't.

        If ``path`` doesn't have a system path,
        a :class:`~fs.errors.NoSysPath` exception will be thrown.

        .. note::

            A filesystem may return a system path even if no
            resource is referenced by that path -- as long as it can
            be certain what that system path would be.

        """
        raise errors.NoSysPath(path=path)

    def gettype(self, path):
        """
        Get the type of a resource.

        :param path: A path in the filesystem.
        :returns: :class:`~fs.ResourceType`

        A type of a resource is an integer that identifies the what
        the resource references. The standard type integers may be one
        of the values in the :class:`~fs.ResourceType` enumerations.

        The most common resource types, supported by virtually all
        filesystems are ``directory`` (1) and ``file`` (2), but the
        following types are also possible:

        ===================   ======
        ResourceType          value
        -------------------   ------
        unknown               0
        directory             1
        file                  2
        character             3
        block_special_file    4
        fifo                  5
        socket                6
        symlink               7
        ===================   ======

        Standard resource types are positive integers, negative values
        are reserved for implementation specific resource types.

        """
        resource_type = self.getdetails(path).type
        return resource_type

    def geturl(self, path, purpose='download'):
        """
        Get a URL to the given resource.

        :param str path: A path on the filesystem
        :param str purpose: A short string that indicates which URL to
            retrieve for the given path (if there is more than one). The
            default is `'download'`, which should return a URL that
            serves the file. Other filesystems may support other values
            for ``purpose``.
        :returns: A URL.
        :rtype: str
        :raises fs.errors.NoURL: If the path does not map to a URL.

        """
        raise errors.NoURL(path, purpose)

    def hassyspath(self, path):
        """
        Check if a path maps to a system path.

        :param str path: A path on the filesystem
        :rtype: bool

        """
        has_sys_path = True
        try:
            self.getsyspath(path)
        except errors.NoSysPath:
            has_sys_path = False
        return has_sys_path

    def hasurl(self, path, purpose='download'):
        """
        Check if a path has a corresponding URL.

        :param str path: A path on the filesystem
        :param str purpose: A purpose parameter, as given in
            :meth:`~fs.base.FS.geturl`.
        :rtype: bool

        """
        has_url = True
        try:
            self.geturl(path, purpose=purpose)
        except errors.NoURL:
            has_url = False
        return has_url

    def isclosed(self):
        """Check if the filesystem is closed."""
        return getattr(self, '_closed', False)

    def isdir(self, path):
        """Check a path exists and is a directory."""
        try:
            return self.getinfo(path).is_dir
        except errors.ResourceNotFound:
            return False

    def isempty(self, path):
        """
        Check if a directory is empty (contains no files or
        directories).

        :param str path: A directory path.
        :rtype: bool

        """
        return next(iter(self.scandir(path)), None) is None

    def isfile(self, path):
        """Check a path exists and is a file."""
        try:
            return not self.getinfo(path).is_dir
        except errors.ResourceNotFound:
            return False

    def islink(self, path):
        """
        Check if a path is a symlink.

        :param str path: A path on the filesystem.
        :rtype: bool

        """
        self.getinfo(path)
        return False

    def lock(self):
        """
        Get a context manager that *locks* the filesystem.

        Locking a filesystem gives a thread exclusive access to it.
        Other threads will block until the threads with the lock has
        left the context manager. Here's how you would use it::

            with my_fs.lock():  # May block
                # code here has exclusive access to the filesystem

        It is a good idea to put a lock around any operations that you
        would like to be *atomic*. For instance if you are copying
        files, and you don't want another thread to delete or modify
        anything while the copy is in progress.

        Locking with this method is only required for code that calls
        multiple filesystem methods. Individual methods are thread safe
        already, and don't need to be locked.

        .. note::
            This only locks at the Python level. There is nothing to
            prevent other processes from modifying the filesystem
            outside of the filesystem instance.

        """
        return self._lock

    def movedir(self, src_path, dst_path, create=False):
        """
        Move contents of directory ``src_path`` to ``dst_path``.

        :param str src_path: Path to source directory on the filesystem.
        :param str dst_path: Path to destination directory.
        :param create: If ``True``, then ``dst_path`` will be created if
            it doesn't already exist.
        :type create: bool

        """
        with self._lock:
            if not create and not self.exists(dst_path):
                raise errors.ResourceNotFound(dst_path)
            move.move_dir(
                self,
                src_path,
                self,
                dst_path
            )

    def makedirs(self, path, permissions=None, recreate=False):
        """
        Make a directory, and any missing intermediate directories.

        :param str path: Path to directory from root.
        :param bool recreate: If ``False`` (default), it is an error to
            attempt to create a directory that already exists. Set to
            `True` to allow directories to be re-created without errors.
        :param permissions: Initial permissions.
        :returns: A sub-directory filesystem.
        :rtype: :class:`~fs.subfs.SubFS`

        :raises fs.errors.DirectoryExists: if the path is already
            a directory, and ``recreate`` is False.
        :raises fs.errors.DirectoryExpected: if one of the ancestors
            in the path isn't a directory.

        """
        self.check()
        with self._lock:
            dir_paths = tools.get_intermediate_dirs(self, path)
            for dir_path in dir_paths:
                self.makedir(dir_path, permissions=permissions)

            try:
                self.makedir(path)
            except errors.DirectoryExists:
                if not recreate:
                    raise
            return self.opendir(path)

    def move(self,
             src_path,
             dst_path,
             overwrite=False):
        """
        Move a file from `src_path` to `dst_path`.

        :param str src_path: A path on the filesystem to move.
        :param str dst_path: A path on the filesystem where the source
            file will be written to.
        :param bool overwrite: If `True` destination path will be
            overwritten if it exists.
        :raises fs.errors.DestinationExists: If ``dst_path`` exists,
            and overwrite == ``False``.
        :raises fs.errors.ResourceNotFound: If a parent directory of
            ``dst_path`` does not exist.

        """

        if not overwrite and self.exists(dst_path):
            raise errors.DestinationExists(dst_path)
        if self.getinfo(src_path).is_dir:
            raise errors.FileExpected(src_path)
        if self.getmeta().get('supports_rename', False):
            try:
                src_sys_path = self.getsyspath(src_path)
                dst_sys_path = self.getsyspath(dst_path)
            except errors.NoSysPath:  # pragma: no cover
                pass
            else:
                try:
                    os.rename(src_sys_path, dst_sys_path)
                except OSError:
                    pass
                else:
                    return
        with self._lock:
            with self.open(src_path, 'rb') as read_file:
                self.setbinfile(dst_path, read_file)
            self.remove(src_path)

    def open(self,
             path,
             mode='r',
             buffering=-1,
             encoding=None,
             errors=None,
             newline='',
             **options):
        """
        Open a file.

        :param str path: A path to a file on the filesystem.
        :param str mode: Mode to open file object.
        :param int buffering: Buffering policy: ``0`` to switch
            buffering off, ``1`` to select line buffering, ``>1`` to
            select a fixed-size buffer, ``-1`` to auto-detect.
        :param str encoding: Encoding for text files (defaults to
            ``utf-8``)
        :param str errors: What to do with unicode decode errors (see
            `stdlib docs <https://docs.python.org/3/library/codecs.html#error-handlers>`_)
        :param str newline: New line parameter (See stdlib docs).
        :param options: Additional keyword parameters to set
            implementation specific options (if required). See
            implementation docs for details.

        :rtype: file object
        """
        validate_open_mode(mode)
        bin_mode = mode.replace('t', '')
        bin_file = self.openbin(path, mode=bin_mode, buffering=buffering)
        io_stream = iotools.make_stream(
            path,
            bin_file,
            mode=mode,
            buffering=buffering,
            encoding=encoding or 'utf-8',
            errors=errors,
            newline=newline,
            **options
        )
        return io_stream

    def opendir(self, path, factory=None):
        """Get a filesystem object for a sub-directory.

        :param str path: Path to a directory on the filesystem.
        :param factory: A callable that when invoked with an FS instance
            and `path` will return a new FS object representing the sub-
            directory contents. If no `factory` is supplied then
            :meth:`~fs.subfs.SubFS` will be used.
        :returns: A filesystem object representing a sub-directory.
        :rtype: :class:`~fs.subfs.SubFS`
        :raises fs.errors.DirectoryExpected: If ``dst_path`` does not
            exist or is not a directory.

        """
        from .subfs import SubFS
        factory = factory or SubFS

        if not self.getbasic(path).is_dir:
            raise errors.DirectoryExpected(
                path=path
            )
        return factory(self, path)

    def removetree(self, dir_path):
        """
        Recursively remove the contents of a directory.

        This method is similar to :meth:`~fs.base.removedir`, but will
        remove the contents of the directory if it is not empty.

        :param str dir_path: Path to a directory on the filesystem.

        """

        _dir_path = abspath(normpath(dir_path))
        with self._lock:
            walker = walk.Walker(search="depth")
            gen_info = walker.info(self, _dir_path)
            for _path, info in gen_info:
                if info.is_dir:
                    self.removedir(_path)
                else:
                    self.remove(_path)
            if _dir_path != '/':
                self.removedir(dir_path)

    def scandir(self, path, namespaces=None, page=None):
        """
        Get an iterator of resource info.

        :param str path: A path on the filesystem
        :param list namespaces: A sequence of info namespaces.
        :param page: May be a tuple of ``(<start>, <end>)`` indexes to
            return an iterator of a subset of the resource info, or
            ``None`` to iterator the entire directory. Paging a
            directory scan may be necessary for very large directories.
        :type page: tuple or None
        :rtype: iterator

        """
        namespaces = namespaces or ()
        _path = abspath(normpath(path))

        info = (
            self.getinfo(
                join(_path, name),
                namespaces=namespaces
            )
            for name in self.listdir(path)
        )
        iter_info = iter(info)
        if page is not None:
            start, end = page
            iter_info = itertools.islice(iter_info, start, end)
        return iter_info

    def setbytes(self, path, contents):
        """
        Copy (bytes) data to a file.

        :param str path: Destination path on the filesystem.
        :param bytes contents: A bytes object with data to be written

        """
        if not isinstance(contents, bytes):
            raise TypeError('contents must be bytes')
        with closing(self.open(path, mode='wb')) as write_file:
            write_file.write(contents)

    def setbinfile(self, path, file):
        """
        Set a file to the contents of a binary file object.

        :param str path: A path on the filesystem.
        :param file: A file object open for reading in binary mode.
        :type file: file object

        This method copies bytes from an open binary file to a file on
        the filesystem. If the destination exists, it will first be
        truncated.

        Note that the file object ``file`` will *not* be closed by this
        method. Take care to close it after this method completes
        (ideally with a context manager). For example::

            with open('myfile.bin') as read_file:
                my_fs.setbinfile('myfile.bin', read_file)

        """

        with self._lock:
            with self.open(path, 'wb') as dst_file:
                tools.copy_file_data(file, dst_file)

    def setfile(self,
                path,
                file,
                encoding=None,
                errors=None,
                newline=None):
        """
        Set a file to the contents of a file object.

        :param str path: A path on the filesystem.
        :param file: A file object open for reading.
        :type file: file object
        :param str encoding: Encoding of destination file, or ``None``
            for binary.
        :param str errors: How encoding errors should be treated (same
            as ``io.open``).
        :param str newline: Newline parameter (same is ``io.open``).

        This method will read the contents of a supplied file object,
        and write to a file on the filesystem. If the destination
        exists, it will first be truncated.

        If `encoding` is supplied, the destination will be opened in
        text mode.

        Note that the file object ``file`` will *not* be closed by this
        method. Take care to close it after this method completes
        (ideally with a context manager). For example::

            with open('myfile.bin') as read_file:
                my_fs.setfile('myfile.bin', read_file)

        """
        mode = 'wb' if encoding is None else 'wt'

        with self._lock:
            with self.open(path,
                           mode=mode,
                           encoding=encoding,
                           errors=errors,
                           newline=newline) as dst_file:
                tools.copy_file_data(file, dst_file)

    def settimes(self, path, accessed=None, modified=None):
        """
        Set the accessed and modified time on a resource.

        :param accessed: The accessed time, as a datetime, or None
            to use the current rime.
        :param modified: The modified time, or ``None`` (the default) to
            use the same time as ``accessed`` parameter.

        """

        details = {}
        raw_info = {
            "details": details
        }

        details['accessed'] = (
            time.time()
            if accessed is None
            else datetime_to_epoch(accessed)
        )

        details['modified'] = (
            details['accessed']
            if modified is None
            else datetime_to_epoch(modified)
        )

        self.setinfo(path, raw_info)

    def settext(self,
                path,
                contents,
                encoding='utf-8',
                errors=None,
                newline=''):
        """
        Create or replace a file with text.

        :param str contents: Path on the filesystem.
        :param str encoding: Encoding of destination file (default
            'UTF-8).
        :param str errors: Error parameter for encoding (same as
            ``io.open``).
        :param str newline: Newline parameter for encoding (same as
            ``io.open``).

        """
        if not isinstance(contents, six.text_type):
            raise TypeError('contents must be unicode')
        with closing(self.open(path,
                               mode="wt",
                               encoding=encoding,
                               errors=errors,
                               newline=newline)) as write_file:
            write_file.write(contents)

    def touch(self, path):
        """
        Create a new file if ``path`` doesn't exist, or update accessed
        and modified times if the path does exist.

        This method is similar to the linux command of the same name.

        :param str path: A path to a file on the filesystem.

        """
        with self._lock:
            now = time.time()
            if not self.create(path):
                raw_info = {
                    "details": {
                        "accessed": now,
                        "modified": now
                    }
                }
                self.setinfo(path, raw_info)

    def validatepath(self, path):
        """
        Check if a path is valid on this filesystem, and return a
        normalized absolute path.

        Many filesystems have restrictions on the format of paths they
        support. This method will check that ``path`` is valid on the
        underlaying storage mechanism and throw a
        :class:`~fs.errors.InvalidPath` exception if it is not.

        :param str path: A path
        :returns: A normalized, absolute path.
        :rtype: str
        :raises fs.errors.InvalidPath: If the path is invalid.
        :raises fs.errors.FilesystemClosed: if the filesystem
            is closed.

        """
        self.check()

        if isinstance(path, bytes):
            raise TypeError(
                'paths must be unicode (not str)'
                if six.PY2 else
                'paths must be str (not bytes)'
            )

        meta = self.getmeta()

        invalid_chars = meta.get('invalid_path_chars')
        if invalid_chars:
            if set(path).intersection(invalid_chars):
                raise errors.InvalidCharsInPath(path)

        max_sys_path_length = meta.get('max_sys_path_length')
        if max_sys_path_length is not None:
            try:
                sys_path = self.getsyspath(path)
            except errors.NoSysPath:  # pragma: no cover
                pass
            else:
                if len(sys_path) > max_sys_path_length:
                    _msg = 'path too long '\
                        '(max {max_chars} characters in sys path)'
                    msg = _msg.format(max_chars=max_sys_path_length)
                    raise errors.InvalidPath(path, msg=msg)
        path = abspath(normpath(path))
        return path

    # ---------------------------------------------------------------- #
    # Helper methods                                                   #
    # Filesystems should not implement these methods.                  #
    # ---------------------------------------------------------------- #

    def getbasic(self, path):
        """
        Get the *basic* resource info.

        :param str path: A path on the filesystem.
        :returns: Resource information object for ``path``.
        :rtype: :class:`~fs.info.Info`

        This method is shorthand for the following::

            fs.getinfo(path, namespaces=['basic'])

        """
        return self.getinfo(path, namespaces=['basic'])

    def getdetails(self, path):
        """
        Get the *details* resource info.

        :param str path: A path on the filesystem.
        :returns: Resource information object for ``path``.
        :rtype: :class:`~fs.info.Info`

        This method is shorthand for the following::

            fs.getinfo(path, namespaces=['details'])

        """
        return self.getinfo(path, namespaces=['details'])

    def check(self):
        """
        Check a filesystem may be used.

        Will throw a :class:`~fs.errors.FilesystemClosed` if the
        filesystem is closed.

        :returns: None
        :raises fs.errors.FilesystemClosed: if the filesystem
            is closed.

        """
        if self.isclosed():
            raise errors.FilesystemClosed()

    def match(self, patterns, name):
        """
        Check if a name matches any of a list of wildcards.

        :param list patterns: A list of patterns, e.g. ``['*.py']``
        :param str name: A file or directory name (not a path)
        :rtype: bool

        If a filesystem is case *insensitive* (such as Windows) then
        this method will perform a case insensitive match (i.e. ``*.py``
        will match the same names as ``*.PY``). Otherwise the match will
        be case sensitive (``*.py`` and ``*.PY`` will match different
        names).

            >>> home_fs.match(['*.py'], '__init__.py')
            True
            >>> home_fs.match(['*.jpg', '*.png'], 'foo.gif')
            False

        If ``patterns`` is ``None``, or (``['*']``), then this method
        will always return True.

        """
        if patterns is None:
            return True
        if isinstance(patterns, six.text_type):
            raise TypeError('patterns must be a list or sequence')
        case_sensitive = self.getmeta().get('case_sensitive', True)
        matcher = wildcard.get_matcher(patterns, case_sensitive)
        return matcher(name)

    def tree(self, **kwargs):
        """
        Render a tree view of the filesystem to stdout or a file.

        The parameters are passed to :func:`~fs.tree.render`.

        """
        from .tree import render
        render(self, **kwargs)



from __future__ import unicode_literals

import io
import os
import mock
import shutil
import tempfile
import unittest

from fs import osfs
from fs.path import relpath
from fs import errors

from fs.test import FSTestCases


from six import text_type


class TestOSFS(FSTestCases, unittest.TestCase):
    """Test OSFS implementation."""

    def make_fs(self):
        temp_dir = tempfile.mkdtemp(u"fstestosfs")
        return osfs.OSFS(temp_dir)

    def destroy_fs(self, fs):
        self.fs.close()

    def _get_real_path(self, path):
        _path = os.path.join(self.fs.root_path, relpath(path))
        return _path

    def assert_exists(self, path):
        _path = self._get_real_path(path)
        self.assertTrue(os.path.exists(_path))

    def assert_not_exists(self, path):
        _path = self._get_real_path(path)
        self.assertFalse(os.path.exists(_path))

    def assert_isfile(self, path):
        _path = self._get_real_path(path)
        self.assertTrue(os.path.isfile(_path))

    def assert_isdir(self, path):
        _path = self._get_real_path(path)
        self.assertTrue(os.path.isdir(_path))

    def assert_bytes(self, path, contents):
        assert isinstance(contents, bytes)
        _path = self._get_real_path(path)
        with io.open(_path, 'rb') as f:
            data = f.read()
        self.assertEqual(data, contents)
        self.assertIsInstance(data, bytes)

    def assert_text(self, path, contents):
        assert isinstance(contents, text_type)
        _path = self._get_real_path(path)
        with io.open(_path, 'rt', encoding='utf-8') as f:
            data = f.read()
        self.assertEqual(data, contents)
        self.assertIsInstance(data, text_type)

    def test_not_exists(self):
        with self.assertRaises(errors.CreateFailed):
            fs = osfs.OSFS('/does/not/exists/')

    def test_create(self):
        """Test create=True"""

        dir_path = tempfile.mkdtemp()
        try:
            create_dir = os.path.join(dir_path, 'test_create')
            with osfs.OSFS(create_dir, create=True):
                self.assertTrue(os.path.isdir(create_dir))
            self.assertTrue(os.path.isdir(create_dir))
        finally:
            shutil.rmtree(dir_path)

        # Test exception when unable to create dir
        with tempfile.NamedTemporaryFile() as tmp_file:
            with self.assertRaises(errors.CreateFailed):
                # Trying to create a dir that exists as a file
                osfs.OSFS(tmp_file.name, create=True)

    def test_unicode_paths(self):
        dir_path = tempfile.mkdtemp()
        try:
            fs_dir = os.path.join(dir_path, u'te\u0161t_\u00fanicod\u0113')
            os.mkdir(fs_dir)
            with osfs.OSFS(fs_dir):
                self.assertTrue(os.path.isdir(fs_dir))
        finally:
            shutil.rmtree(dir_path)

    def test_symlinks(self):
        with open(self._get_real_path('foo'), 'wb') as f:
            f.write(b'foobar')
        os.symlink(self._get_real_path('foo'), self._get_real_path('bar'))
        self.assertFalse(self.fs.islink('foo'))
        self.assertFalse(self.fs.getinfo('foo', namespaces=['link']).is_link)
        self.assertTrue(self.fs.islink('bar'))
        self.assertTrue(self.fs.getinfo('bar', namespaces=['link']).is_link)

        foo_info = self.fs.getinfo('foo', namespaces=['link', 'lstat'])
        self.assertIn('link', foo_info.raw)
        self.assertIn('lstat', foo_info.raw)
        self.assertEqual(foo_info.get('link', 'target'), None)
        self.assertEqual(foo_info.target, foo_info.raw['link']['target'])
        bar_info = self.fs.getinfo('bar', namespaces=['link', 'lstat'])
        self.assertIn('link', bar_info.raw)
        self.assertIn('lstat', bar_info.raw)




from __future__ import print_function
from __future__ import unicode_literals

from collections import OrderedDict
import calendar
import io
import itertools
import socket
import threading

from contextlib import contextmanager
from ftplib import FTP
from ftplib import error_perm
from ftplib import error_temp

from six import PY2
from six import text_type

from . import errors
from .base import FS
from .constants import DEFAULT_CHUNK_SIZE
from .enums import ResourceType
from .enums import Seek
from .info import Info
from .iotools import line_iterator
from .mode import Mode
from .path import abspath
from .path import basename
from .path import normpath
from .path import split
from . import _ftp_parse as ftp_parse


__all__ = ['FTPFS']


@contextmanager
def ftp_errors(fs, path=None):
    try:
        with fs._lock:
            yield
    except socket.error:
        raise errors.RemoteConnectionError(
            msg='unable to connect to {}'.format(fs.host)
        )
    except error_temp as error:
        if path is not None:
            raise errors.ResourceError(
                path,
                msg="ftp error on resource '{}' ({})".format(path, error)
            )
        else:
            raise errors.OperationFailed(
                msg='ftp error ({})'.format(error)
            )
    except error_perm as error:
        code, message = _parse_ftp_error(error)
        if code == 552:
            raise errors.InsufficientStorage(
                path=path,
                msg=message
            )
        elif code in (501, 550):
            raise errors.ResourceNotFound(path=path)
        raise errors.PermissionDenied(
            msg=message
        )


@contextmanager
def manage_ftp(ftp):
    try:
        yield ftp
    finally:
        try:
            ftp.quit()
        except:  # pragma: nocover
            pass


def _parse_ftp_error(error):
    """Extract code and message from ftp error."""
    code, _, message = text_type(error).partition(' ')
    if code.isdigit():
        code = int(code)
    return code, message


if PY2:
    def _encode(st, encoding):
        return st.encode(encoding) if isinstance(st, text_type) else st
    def _decode(st, encoding):
        return st.decode(encoding, 'replace') if isinstance(st, bytes) else st
else:
    def _encode(st, _):
        return st
    def _decode(st, _):
        return st


class FTPFile(io.IOBase):

    def __init__(self, ftpfs, path, mode):
        super(FTPFile, self).__init__()
        self.fs = ftpfs
        self.path = path
        self.mode = Mode(mode)
        self.pos = 0
        self._lock = threading.Lock()
        self.ftp = self._open_ftp()
        self._read_conn = None
        self._write_conn = None

    def _open_ftp(self):
        """Open an ftp object for the file."""
        ftp = self.fs._open_ftp()
        ftp.voidcmd(str('TYPE I'))
        return ftp

    @property
    def read_conn(self):
        if self._read_conn is None:
            self._read_conn = self.ftp.transfercmd(
                str('RETR ') + _encode(self.path, self.ftp.encoding),
                self.pos
            )
        return self._read_conn

    @property
    def write_conn(self):
        if self._write_conn is None:
            if self.mode.appending:
                self._write_conn = self.ftp.transfercmd(
                    str('APPE ') + _encode(self.path, self.ftp.encoding)
                )
            else:
                self._write_conn = self.ftp.transfercmd(
                    str('STOR ') + _encode(self.path, self.ftp.encoding),
                    self.pos
                )
        return self._write_conn

    def __repr__(self):
        _repr = "<ftpfile {!r} {!r} {!r}>"
        return _repr.format(self.fs.ftp_url, self.path, self.mode)

    def close(self):
        if not self.closed:
            with self._lock:
                try:
                    if self._write_conn is not None:
                        self._write_conn.close()
                        self._write_conn = None
                    if self._read_conn is not None:
                        self._read_conn.close()
                        self._read_conn = None
                    try:
                        self.ftp.quit()
                    except error_temp:  # pragma: nocover
                        pass
                finally:
                    super(FTPFile, self).close()

    def tell(self):
        return self.pos

    def readable(self):
        return self.mode.reading

    def read(self, size=-1):
        if not self.mode.reading:
            raise IOError('File not open for reading')

        chunks = []
        remaining = size

        conn = self.read_conn
        with self._lock:
            while remaining:
                if remaining < 0:
                    read_size = DEFAULT_CHUNK_SIZE
                else:
                    read_size = min(DEFAULT_CHUNK_SIZE, remaining)
                try:
                    chunk = conn.recv(read_size)
                except socket.error:  # pragma: nocover
                    break
                if not chunk:
                    break
                chunks.append(chunk)
                self.pos += len(chunk)
                remaining -= len(chunk)
        return b''.join(chunks)

    def readline(self, size=None):
        return next(line_iterator(self, size))

    def readlines(self, hint=-1):
        lines = []
        size = 0
        for line in line_iterator(self):
            lines.append(line)
            size += len(line)
            if hint != -1 and size > hint:
                break
        return lines

    def writable(self):
        return self.mode.writing

    def write(self, data):
        if not self.mode.writing:
            raise IOError('File not open for writing')

        with self._lock:
            conn = self.write_conn
            data_pos = 0
            remaining_data = len(data)

            while remaining_data:
                chunk_size = min(remaining_data, DEFAULT_CHUNK_SIZE)
                sent_size = conn.send(data[data_pos:data_pos + chunk_size])
                data_pos += sent_size
                remaining_data -= sent_size
                self.pos += sent_size

        return data_pos

    def writelines(self, lines):
        self.write(b''.join(lines))

    def truncate(self, size=None):
        # Inefficient, but I don't know if truncate is possible with ftp
        with self._lock:
            if size is None:
                size = self.tell()
            with self.fs.openbin(self.path) as f:
                data = f.read(size)
            with self.fs.openbin(self.path, 'w') as f:
                f.write(data)
                if len(data) < size:
                    f.write(b'\0' * (size - len(data)))
        return size

    def seekable(self):
        return True

    def seek(self, pos, whence=Seek.set):
        if whence not in (Seek.set, Seek.current, Seek.end):
            raise ValueError('invalid value for whence')
        with self._lock:
            if whence == Seek.set:
                new_pos = pos
            elif whence == Seek.current:
                new_pos = self.pos + pos
            elif whence == Seek.end:
                file_size = self.fs.getsize(self.path)
                new_pos = file_size + pos
            self.pos = max(0, new_pos)

            self.ftp.quit()
            self.ftp = self._open_ftp()

            if self._read_conn:
                self._read_conn.close()
                self._read_conn = None
            if self._write_conn:
                self._write_conn.close()
                self._write_conn = None
        return self.tell()


class FTPFS(FS):
    """
    A FTP (File Transport Protocol) Filesystem.

    :param str host: A FTP host, e.g. ``'ftp.mirror.nl'``.
    :param str user: A username (default is ``'anonymous'``)
    :param passwd: Password for the server, or ``None`` for anon.
    :param acct: FTP account.
    :param int timeout: Timeout for contacting server (in seconds).
    :param int port: Port number (default 21).

    """

    _meta = {
        'invalid_path_chars': '\0',
        'network': True,
        'read_only': False,
        'thread_safe': True,
        'unicode_paths': True,
        'virtual': False,
    }

    def __init__(self,
                 host,
                 user='anonymous',
                 passwd='',
                 acct='',
                 timeout=10,
                 port=21):
        super(FTPFS, self).__init__()
        self.host = host
        self.user = user
        self.passwd = passwd
        self.acct = acct
        self.timeout = timeout
        self.port = port

        self.encoding = 'latin-1'
        self._ftp = None
        self._welcome = None
        self._features = None

    def __repr__(self):
        return "FTPFS({!r}, port={!r})".format(self.host, self.port)

    def __str__(self):
        _fmt = (
            "<ftpfs '{host}'>"
            if self.port == 21
            else "<ftpfs '{host}:{port}'>"
        )
        return _fmt.format(host=self.host, port=self.port)

    @classmethod
    def _parse_features(cls, feat_response):
        """Parse a dict of features from FTP feat response."""
        features = {}
        if feat_response.split('-')[0] == '211':
            for line in feat_response.splitlines():
                if line.startswith(' '):
                    key, _, value = line[1:].partition(' ')
                    features[key] = value
        return features

    def _open_ftp(self):
        """Open a new ftp object."""
        _ftp = FTP()
        _ftp.set_debuglevel(0)
        with ftp_errors(self):
            _ftp.connect(self.host, self.port, self.timeout)
            _ftp.login(self.user, self.passwd, self.acct)
            self._features = {}
            try:
                feat_response = _decode(_ftp.sendcmd("FEAT"), 'latin-1')
            except error_perm:
                self.encoding = 'latin-1'
            else:
                self._features = self._parse_features(feat_response)
                self.encoding = (
                    'utf-8'
                    if 'UTF8' in self._features
                    else 'latin-1'
                )
                if not PY2:
                    _ftp.file = _ftp.sock.makefile(
                        'r',
                        encoding=self.encoding
                    )
        _ftp.encoding = self.encoding
        self._welcome = _ftp.welcome
        return _ftp

    def _manage_ftp(self):
        ftp = self._open_ftp()
        return manage_ftp(ftp)

    @property
    def ftp_url(self):
        """Get the FTP url this filesystem will open."""
        url = (
            "ftp://{}".format(self.host)
            if self.port == 21
            else "ftp://{}:{}".format(self.host, self.port)
        )
        return url

    @property
    def ftp(self):
        """Get a FTP (ftplib) object."""
        return self._get_ftp()

    def _get_ftp(self):
        if self._ftp is None:
            self._ftp = self._open_ftp()
        return self._ftp

    @property
    def features(self):
        """Get features dict from FTP server."""
        self._get_ftp()
        return self._features

    def _read_dir(self, path):
        _path = abspath(normpath(path))
        lines = []
        with ftp_errors(self, path=path):
            self.ftp.retrlines(
                str('LIST ') + _encode(_path, self.ftp.encoding),
                lines.append
            )
        lines = [
            line.decode('utf-8') if isinstance(line, bytes) else line
            for line in lines
        ]
        _list = [Info(raw_info) for raw_info in ftp_parse.parse(lines)]
        dir_listing = OrderedDict({info.name: info for info in _list})
        return dir_listing

    @property
    def supports_mlst(self):
        """Check if server supports MLST feature."""
        return 'MLST' in self.features

    def create(self, path, wipe=False):
        _path = self.validatepath(path)
        with ftp_errors(self, path):
            if wipe or not self.isfile(path):
                empty_file = io.BytesIO()
                self.ftp.storbinary(
                    str("STOR ") + _encode(_path, self.ftp.encoding),
                    empty_file
                )

    @classmethod
    def _parse_ftp_time(cls, time_text):
        """Parse a time from an ftp directory listing."""
        try:
            tm_year = int(time_text[0:4])
            tm_month = int(time_text[4:6])
            tm_day = int(time_text[6:8])
            tm_hour = int(time_text[8:10])
            tm_min = int(time_text[10:12])
            tm_sec = int(time_text[12:14])
        except ValueError:
            return None
        epoch_time = calendar.timegm((
            tm_year,
            tm_month,
            tm_day,
            tm_hour,
            tm_min,
            tm_sec
        ))
        return epoch_time

    @classmethod
    def _parse_facts(cls, line):
        name = None
        facts = {}
        for fact in line.split(';'):
            key, sep, value = fact.partition('=')
            if sep:
                key = key.strip().lower()
                value = value.strip()
                facts[key] = value
            else:
                name = basename(fact.rstrip('/').strip())
        return name if name not in ('.', '..') else None, facts

    @classmethod
    def _parse_mlsx(cls, lines):
        for line in lines:
            name, facts = cls._parse_facts(line.strip())
            if name is None:
                continue
            is_dir = facts.get('type', None) in ('dir', 'cdir', 'pdir')
            raw_info = {
                "basic":
                {
                    "name": name,
                    "is_dir": is_dir,
                },
                "details":
                {
                    "type": (
                        int(ResourceType.directory)
                        if is_dir else
                        int(ResourceType.file)
                    )
                },
                "ftp": facts
            }
            details = raw_info['details']
            size_str = facts.get('size', facts.get('sizd', '0'))
            size = 0
            if size_str.isdigit():
                size = int(size_str)
            details['size'] = size
            if 'modify' in facts:
                details['modified'] = cls._parse_ftp_time(facts['modify'])
            if 'create' in facts:
                details['created'] = cls._parse_ftp_time(facts['create'])
            yield raw_info

    def getinfo(self, path, namespaces=None):
        _path = self.validatepath(path)
        namespaces = namespaces or ()

        if _path == '/':
            return Info({
                "basic":
                {
                    "name": "",
                    "is_dir": True
                },
                "details":
                {
                    "type": int(ResourceType.directory)
                }
            })

        if self.supports_mlst:
            with self._lock:
                with ftp_errors(self, path=path):
                    response = self.ftp.sendcmd(
                        str('MLST ') + _encode(_path, self.ftp.encoding)
                    )
                lines = _decode(response, self.ftp.encoding).splitlines()[1:-1]
                for raw_info in self._parse_mlsx(lines):
                    return Info(raw_info)

        with ftp_errors(self, path=path):
            dir_name, file_name = split(_path)
            directory = self._read_dir(dir_name)
            if file_name not in directory:
                raise errors.ResourceNotFound(path)
            info = directory[file_name]
            return info

    def getmeta(self, namespace="standard"):
        _meta = {}
        self._get_ftp()
        if namespace == "standard":
            _meta = self._meta.copy()
            _meta['unicode_paths'] = "UTF8" in self.features
        return _meta


    def listdir(self, path):
        _path = self.validatepath(path)
        with self._lock:
            dir_list = [
                info.name
                for info in self.scandir(_path)
            ]
        return dir_list

    def makedir(self, path, permissions=None, recreate=False):
        _path = self.validatepath(path)

        with ftp_errors(self, path=path):
            if _path == '/':
                if recreate:
                    return self.opendir(path)
                else:
                    raise errors.DirectoryExists(path)

            if not (recreate and self.isdir(path)):
                try:
                    self.ftp.mkd(_encode(_path, self.ftp.encoding))
                except error_perm as error:
                    code, _ = _parse_ftp_error(error)
                    if code == 550:
                        if self.isdir(path):
                            raise errors.DirectoryExists(path)
                        else:
                            if self.exists(path):
                                raise errors.DirectoryExists(path)
                    raise errors.ResourceNotFound(path)
        return self.opendir(path)

    def openbin(self, path, mode="r", buffering=-1, **options):
        _mode = Mode(mode)
        _mode.validate_bin()
        _path = self.validatepath(path)

        with self._lock:
            try:
                info = self.getinfo(_path)
            except errors.ResourceNotFound:
                if _mode.reading:
                    raise errors.ResourceNotFound(path)
            else:
                if info.is_dir:
                    raise errors.FileExpected(path)
            if _mode.exclusive:
                raise errors.FileExists(path)
            ftp_file = FTPFile(self, _path, mode)
        return ftp_file

    def remove(self, path):
        self.check()
        _path = self.validatepath(path)
        with self._lock:
            if self.isdir(path):
                raise errors.FileExpected(path=path)
            with ftp_errors(self, path):
                self.ftp.delete(_encode(_path, self.ftp.encoding))

    def removedir(self, path):
        _path = self.validatepath(path)
        if _path == '/':
            raise errors.RemoveRootError()

        with ftp_errors(self, path):
            try:
                self.ftp.rmd(_encode(_path, self.ftp.encoding))
            except error_perm as error:
                code, _ = _parse_ftp_error(error)
                if code == 550:
                    if self.isfile(path):
                        raise errors.DirectoryExpected(path)
                    if not self.isempty(path):
                        raise errors.DirectoryNotEmpty(path)
                raise  # pragma: no cover

    def _scandir(self, path, namespaces=None):
        _path = self.validatepath(path)
        with self._lock:
            if self.supports_mlst:
                lines = []
                with ftp_errors(self, path=path):
                    try:
                        self.ftp.retrlines(
                            str("MLSD ") + _encode(_path, self.ftp.encoding),
                            lambda l: lines.append(_decode(l, self.ftp.encoding))
                        )
                    except error_perm:
                        if not self.getinfo(path).is_dir:
                            raise errors.DirectoryExpected(path)
                        raise # pragma: no cover
                if lines:
                    for raw_info in self._parse_mlsx(lines):
                        yield Info(raw_info)
                    return
            with self._lock:
                for info in self._read_dir(_path).values():
                    yield info

    def scandir(self, path, namespaces=None, page=None):
        if not self.supports_mlst and not self.getinfo(path).is_dir:
            raise errors.DirectoryExpected(path)
        iter_info = self._scandir(path, namespaces=namespaces)
        if page is not None:
            start, end = page
            iter_info = itertools.islice(iter_info, start, end)
        return iter_info

    def setbinfile(self, path, file):
        _path = self.validatepath(path)
        with self._lock:
            with self._manage_ftp() as ftp:
                with ftp_errors(self, path):
                    ftp.storbinary(
                        str("STOR ") + _encode(_path, self.ftp.encoding),
                        file
                    )

    def setbytes(self, path, contents):
        if not isinstance(contents, bytes):
            raise TypeError('contents must be bytes')
        self.setbinfile(path, io.BytesIO(contents))

    def setinfo(self, path, info):
        if not self.exists(path):
            raise errors.ResourceNotFound(path)

    def getbytes(self, path):
        _path = self.validatepath(path)
        data = io.BytesIO()
        with ftp_errors(self, path):
            with self._manage_ftp() as ftp:
                try:
                    ftp.retrbinary(
                        str("RETR ") + _encode(_path, self.ftp.encoding),
                        data.write
                    )
                except error_perm as error:
                    code, _ = _parse_ftp_error(error)
                    if code == 550:
                        if self.isdir(path):
                            raise errors.FileExpected(path)
                    raise

        data_bytes = data.getvalue()
        return data_bytes

    def close(self):
        if not self.isclosed():
            try:
                self.ftp.quit()
            except Exception:  # pragma: no cover
                pass
            self._ftp = None
        super(FTPFS, self).close()



from __future__ import unicode_literals

import unittest

from fs import lrucache


class TestLRUCache(unittest.TestCase):

    def setUp(self):
        self.lrucache = lrucache.LRUCache(3)

    def test_lrucache(self):
        # insert some values
        self.lrucache['foo'] = 1
        self.lrucache['bar'] = 2
        self.lrucache['baz'] = 3
        self.assertIn('foo', self.lrucache)

        #  Cache size is 3, so the following should kick oldest one out
        self.lrucache['egg'] = 4
        self.assertNotIn('foo', self.lrucache)
        self.assertIn('egg', self.lrucache)

        # cache is now full
        # look up two keys
        self.lrucache['bar']
        self.lrucache['baz']

        # Insert a new value
        self.lrucache['eggegg'] = 5
        # Check it kicked out the 'oldest' key
        self.assertNotIn('egg', self.lrucache)



# coding: utf-8
"""
fs.opener.base
==============

Defines the Opener abstract base class.
"""

import six
import abc


@six.add_metaclass(abc.ABCMeta)
class Opener(object):
    """
    The opener base class.

    An opener is responsible for opening a filesystem for a given
    protocol.

    """

    protocols = []

    def __repr__(self):
        return "<opener {!r}>".format(self.protocols)

    @abc.abstractmethod
    def open_fs(self, fs_url, parse_result, writeable, create, cwd):
        """
        Open a filesystem object from a FS URL.

        :param str fs_url: A filesystem URL
        :param parse_result: A parsed filesystem URL.
        :type parse_result: :class:`~fs.opener.parse.ParseResult`
        :param bool writeable: True if the filesystem must be writeable.
        :param bool create: True if the filesystem should be created if
            it does not exist.
        :param str cwd: The current working directory (generally only
            relevant for OS filesystems).
        :returns: :class:`~fs.base.FS` object

        """



from __future__ import unicode_literals

import io
import unittest

from fs import tree
from fs.memoryfs import MemoryFS


class TestInfo(unittest.TestCase):

    def setUp(self):
        self.fs = MemoryFS()
        self.fs.makedir('foo')
        self.fs.makedir('bar')
        self.fs.makedir('baz')
        self.fs.makedirs('foo/egg1')
        self.fs.makedirs('foo/egg2')
        self.fs.create('/root1')
        self.fs.create('/root2')
        self.fs.create('/foo/test.txt')
        self.fs.create('/foo/test2.txt')
        self.fs.create('/foo/.hidden')
        self.fs.makedirs('/deep/deep1/deep2/deep3/deep4/deep5/deep6')

    def test_tree(self):

        output_file = io.StringIO()

        tree.render(self.fs, file=output_file)

        expected = '|-- bar\n|-- baz\n|-- deep\n|   `-- deep1\n|       `-- deep2\n|           `-- deep3\n|               `-- deep4\n|                   `-- deep5\n|-- foo\n|   |-- egg1\n|   |-- egg2\n|   |-- .hidden\n|   |-- test.txt\n|   `-- test2.txt\n|-- root1\n`-- root2\n'
        self.assertEqual(output_file.getvalue(), expected)

    def test_tree_encoding(self):

        output_file = io.StringIO()

        tree.render(self.fs, file=output_file, with_color=True)

        print(repr(output_file.getvalue()))

        expected = u'\x1b[32m\u251c\u2500\u2500\x1b[0m \x1b[1;34mbar\x1b[0m\n\x1b[32m\u251c\u2500\u2500\x1b[0m \x1b[1;34mbaz\x1b[0m\n\x1b[32m\u251c\u2500\u2500\x1b[0m \x1b[1;34mdeep\x1b[0m\n\x1b[32m\u2502   \u2514\u2500\u2500\x1b[0m \x1b[1;34mdeep1\x1b[0m\n\x1b[32m\u2502       \u2514\u2500\u2500\x1b[0m \x1b[1;34mdeep2\x1b[0m\n\x1b[32m\u2502           \u2514\u2500\u2500\x1b[0m \x1b[1;34mdeep3\x1b[0m\n\x1b[32m\u2502               \u2514\u2500\u2500\x1b[0m \x1b[1;34mdeep4\x1b[0m\n\x1b[32m\u2502                   \u2514\u2500\u2500\x1b[0m \x1b[1;34mdeep5\x1b[0m\n\x1b[32m\u251c\u2500\u2500\x1b[0m \x1b[1;34mfoo\x1b[0m\n\x1b[32m\u2502   \u251c\u2500\u2500\x1b[0m \x1b[1;34megg1\x1b[0m\n\x1b[32m\u2502   \u251c\u2500\u2500\x1b[0m \x1b[1;34megg2\x1b[0m\n\x1b[32m\u2502   \u251c\u2500\u2500\x1b[0m \x1b[33m.hidden\x1b[0m\n\x1b[32m\u2502   \u251c\u2500\u2500\x1b[0m test.txt\n\x1b[32m\u2502   \u2514\u2500\u2500\x1b[0m test2.txt\n\x1b[32m\u251c\u2500\u2500\x1b[0m root1\n\x1b[32m\u2514\u2500\u2500\x1b[0m root2\n'
        self.assertEqual(output_file.getvalue(), expected)

    def test_tree_bytes_no_dirs_first(self):

        output_file = io.StringIO()

        tree.render(self.fs, file=output_file, dirs_first=False)

        expected = u'|-- bar\n|-- baz\n|-- deep\n|   `-- deep1\n|       `-- deep2\n|           `-- deep3\n|               `-- deep4\n|                   `-- deep5\n|-- foo\n|   |-- .hidden\n|   |-- egg1\n|   |-- egg2\n|   |-- test.txt\n|   `-- test2.txt\n|-- root1\n`-- root2\n'
        self.assertEqual(output_file.getvalue(), expected)

    def test_error(self):
        output_file = io.StringIO()

        filterdir = self.fs.filterdir

        def broken_filterdir(path, **kwargs):
            if path.startswith('/deep/deep1/'):
                # Because error messages differ accross Python versions
                raise Exception('integer division or modulo by zero')
            return filterdir(path, **kwargs)
        self.fs.filterdir = broken_filterdir
        tree.render(self.fs, file=output_file, with_color=True)

        expected = u'\x1b[32m\u251c\u2500\u2500\x1b[0m \x1b[1;34mbar\x1b[0m\n\x1b[32m\u251c\u2500\u2500\x1b[0m \x1b[1;34mbaz\x1b[0m\n\x1b[32m\u251c\u2500\u2500\x1b[0m \x1b[1;34mdeep\x1b[0m\n\x1b[32m\u2502   \u2514\u2500\u2500\x1b[0m \x1b[1;34mdeep1\x1b[0m\n\x1b[32m\u2502       \u2514\u2500\u2500\x1b[0m \x1b[1;34mdeep2\x1b[0m\n\x1b[32m\u2502           \u2514\u2500\u2500\x1b[0m \x1b[31merror (integer division or modulo by zero)\x1b[0m\n\x1b[32m\u251c\u2500\u2500\x1b[0m \x1b[1;34mfoo\x1b[0m\n\x1b[32m\u2502   \u251c\u2500\u2500\x1b[0m \x1b[1;34megg1\x1b[0m\n\x1b[32m\u2502   \u251c\u2500\u2500\x1b[0m \x1b[1;34megg2\x1b[0m\n\x1b[32m\u2502   \u251c\u2500\u2500\x1b[0m \x1b[33m.hidden\x1b[0m\n\x1b[32m\u2502   \u251c\u2500\u2500\x1b[0m test.txt\n\x1b[32m\u2502   \u2514\u2500\u2500\x1b[0m test2.txt\n\x1b[32m\u251c\u2500\u2500\x1b[0m root1\n\x1b[32m\u2514\u2500\u2500\x1b[0m root2\n'
        tree_output = output_file.getvalue()
        print(repr(tree_output))

        self.assertEqual(expected, tree_output)

        output_file = io.StringIO()
        tree.render(self.fs, file=output_file, with_color=False)

        expected = u'|-- bar\n|-- baz\n|-- deep\n|   `-- deep1\n|       `-- deep2\n|           `-- error (integer division or modulo by zero)\n|-- foo\n|   |-- egg1\n|   |-- egg2\n|   |-- .hidden\n|   |-- test.txt\n|   `-- test2.txt\n|-- root1\n`-- root2\n'
        self.assertEqual(expected, output_file.getvalue())



from __future__ import print_function
from __future__ import unicode_literals

import io
from io import SEEK_SET, SEEK_CUR

from .mode import Mode


class RawWrapper(io.IOBase):
    """Convert a Python 2 style file-like object in to a IO object."""

    def __init__(self, f, mode=None, name=None):
        self._f = f
        self.mode = mode or getattr(f, 'mode', None)
        self.name = name
        super(RawWrapper, self).__init__()

    def close(self):
        if not self.closed:
            # Close self first since it will
            # flush itself, so we can't close
            # self._f before that
            super(RawWrapper, self).close()
            self._f.close()

    def fileno(self):
        return self._f.fileno()

    def flush(self):
        return self._f.flush()

    def isatty(self):
        return self._f.isatty()

    def seek(self, offset, whence=SEEK_SET):
        return self._f.seek(offset, whence)

    def readable(self):
        return getattr(
            self._f,
            'readable',
            lambda: Mode(self.mode).reading
        )()

    def writable(self):
        return getattr(
            self._f,
            'writable',
            lambda: Mode(self.mode).writing
        )()

    def seekable(self):
        try:
            return self._f.seekable()
        except AttributeError:
            try:
                self.seek(0, SEEK_CUR)
            except IOError:
                return False
            else:
                return True

    def tell(self):
        return self._f.tell()

    def truncate(self, size=None):
        return self._f.truncate(size)

    def write(self, data):
        count = self._f.write(data)
        return len(data) if count is None else count

    def read(self, n=-1):
        if n == -1:
            return self.readall()
        return self._f.read(n)

    def read1(self, n=-1):
        if hasattr(self._f, 'read1'):
            return self._f.read1(n)
        return self.read(n)

    def readall(self):
        return self._f.read()

    def readinto(self, b):
        try:
            return self._f.readinto(b)
        except AttributeError:
            data = self._f.read(len(b))
            bytes_read = len(data)
            b[:len(data)] = data
            return bytes_read

    def readinto1(self, b):
        try:
            return self._f.readinto1(b)
        except AttributeError:
            data = self._f.read1(len(b))
            bytes_read = len(data)
            b[:len(data)] = data
            return bytes_read

    def readline(self, limit=-1):
        return self._f.readline(limit)

    def readlines(self, hint=-1):
        return self._f.readlines(hint)

    def writelines(self, sequence):
        return self._f.writelines(sequence)

    def __iter__(self):
        return iter(self._f)


def make_stream(name,
                bin_file,
                mode='r',
                buffering=-1,
                encoding=None,
                errors=None,
                newline='',
                line_buffering=False,
                **kwargs):
    """Take a Python 2.x binary file and return an IO Stream."""
    reading = 'r' in mode
    writing = 'w' in mode
    appending = 'a' in mode
    binary = 'b' in mode
    if '+' in mode:
        reading = True
        writing = True

    encoding = None if binary else (encoding or 'utf-8')

    io_object = RawWrapper(bin_file, mode=mode, name=name)
    if buffering >= 0:
        if reading and writing:
            io_object = io.BufferedRandom(
                io_object,
                buffering or io.DEFAULT_BUFFER_SIZE
            )
        elif reading:
            io_object = io.BufferedReader(
                io_object,
                buffering or io.DEFAULT_BUFFER_SIZE
            )
        elif writing or appending:
            io_object = io.BufferedWriter(
                io_object,
                buffering or io.DEFAULT_BUFFER_SIZE
            )

    if not binary:
        io_object = io.TextIOWrapper(
            io_object,
            encoding=encoding,
            errors=errors,
            newline=newline,
            line_buffering=line_buffering,
        )

    return io_object


def line_iterator(readable_file, size=None):
    """A not terribly efficient char by char line iterator."""
    read = readable_file.read
    line = []
    byte = b'1'
    if size is None or size < 0:
        while byte:
            byte = read(1)
            line.append(byte)
            if byte in b'\n':
                yield b''.join(line)
                del line[:]

    else:
        while byte and size:
            byte = read(1)
            size -= len(byte)
            line.append(byte)
            if byte in b'\n' or not size:
                yield b''.join(line)
                del line[:]



from __future__ import unicode_literals

import unittest

import six

from fs.appfs import UserDataFS



class TestAppFS(unittest.TestCase):
    """Test Application FS."""

    def test_user_data(self):
        """Test UserDataFS."""
        user_data_fs = UserDataFS('fstest', 'willmcgugan', '1.0')
        if six.PY2:
            self.assertEqual(
                repr(user_data_fs),
                "UserDataFS(u'fstest', author=u'willmcgugan', version=u'1.0')"
            )
        else:
            self.assertEqual(
                repr(user_data_fs),
                "UserDataFS('fstest', author='willmcgugan', version='1.0')"
            )
        self.assertEqual(
            str(user_data_fs),
            "<userdatafs 'fstest'>"
        )



# coding: utf-8
"""Defines the FTPOpener."""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from .base import Opener

class FTPOpener(Opener):
    protocols = ['ftp']

    def open_fs(self, fs_url, parse_result, writeable, create, cwd):
        from ..ftpfs import FTPFS
        from ..subfs import ClosingSubFS
        ftp_host, _, dir_path = parse_result.resource.partition('/')
        ftp_host, _, ftp_port = ftp_host.partition(':')
        ftp_port = int(ftp_port) if ftp_port.isdigit() else 21
        ftp_fs = FTPFS(
            ftp_host,
            port=ftp_port,
            user=parse_result.username,
            passwd=parse_result.password,
        )
        ftp_fs = (
            ftp_fs.opendir(dir_path, factory=ClosingSubFS)
            if dir_path else
            ftp_fs
        )
        return ftp_fs



"""
A collection of functions that operate on filesystems and tools.

"""

from __future__ import print_function
from __future__ import unicode_literals

import io

from . import errors
from .errors import DirectoryNotEmpty
from .errors import ResourceNotFound
from .path import abspath
from .path import dirname
from .path import normpath
from .path import recursepath


def remove_empty(fs, path):
    """
    Remove all empty parents.

    :param fs: A FS object.
    :param str path: Path to a directory on the filesystem.

    """
    path = abspath(normpath(path))
    try:
        while path not in ('', '/'):
            fs.removedir(path)
            path = dirname(path)
    except DirectoryNotEmpty:
        pass


def copy_file_data(src_file, dst_file, chunk_size=None):
    """
    Copy data from one file object to another.

    :param file-like src_file: File open for reading.
    :param file-like dst_file: File open for writing.
    :param int chunk_size: Number of bytes to copy at a time (or
        ``None`` to use sensible default).

    """
    chunk_size = chunk_size or io.DEFAULT_BUFFER_SIZE
    read = src_file.read
    write = dst_file.write
    # The 'or None' is so that it works with binary and text files
    for chunk in iter(lambda: read(chunk_size) or None, None):
        write(chunk)


def get_intermediate_dirs(fs, dir_path):
    """
    Get paths of intermediate directories required to create a new
    directory.

    :param fs: A FS object.
    :param str dir_path: A path to a new directory on the filesystem.
    :returns: A list of paths.
    :rtype: list

    :raises `fs.errors.DirectoryExpected`: If a path component
        references a file and not a directory.

    """
    intermediates = []
    with fs.lock():
        for path in recursepath(abspath(dir_path), reverse=True):
            try:
                resource = fs.getinfo(path)
            except ResourceNotFound:
                intermediates.append(abspath(path))
            else:
                if resource.is_dir:
                    break
                raise errors.DirectoryExpected(dir_path)
    return intermediates[::-1][:-1]



from __future__ import unicode_literals

import os
import mock
import tempfile
import unittest
import pkg_resources

import six

from fs import open_fs, opener
from fs.osfs import OSFS
from fs.opener import registry, errors
from fs.memoryfs import MemoryFS
from fs.appfs import UserDataFS
from fs.opener.parse import ParseResult


class TestParse(unittest.TestCase):

    def test_registry_repr(self):
        str(registry)
        repr(registry)

    def test_parse_not_url(self):
        with self.assertRaises(errors.ParseError):
            parsed = opener.parse('foo/bar')

    def test_parse_simple(self):
        parsed = opener.parse('osfs://foo/bar')
        expected = ParseResult(
            'osfs',
            None,
            None,
            'foo/bar',
            {},
            None
        )
        self.assertEqual(expected, parsed)

    def test_parse_credentials(self):
        parsed = opener.parse('ftp://user:pass@ftp.example.org')
        expected = ParseResult(
            'ftp',
            'user',
            'pass',
            'ftp.example.org',
            {},
            None
        )
        self.assertEqual(expected, parsed)

        parsed = opener.parse('ftp://user@ftp.example.org')
        expected = ParseResult(
            'ftp',
            'user',
            '',
            'ftp.example.org',
            {},
            None
        )
        self.assertEqual(expected, parsed)

    def test_parse_path(self):
        parsed = opener.parse('osfs://foo/bar!example.txt')
        expected = ParseResult(
            'osfs',
            None,
            None,
            'foo/bar',
            {},
            'example.txt'
        )
        self.assertEqual(expected, parsed)

    def test_parse_params(self):
        parsed = opener.parse('ftp://ftp.example.org?proxy=ftp.proxy.org')
        expected = ParseResult(
            'ftp',
            None,
            None,
            'ftp.example.org',
            {
                'proxy':'ftp.proxy.org'
            },
            None
        )
        self.assertEqual(expected, parsed)

    def test_parse_params_multiple(self):
        parsed = opener.parse('ftp://ftp.example.org?foo&bar=1')
        expected = ParseResult(
            'ftp',
            None,
            None,
            'ftp.example.org',
            {
                'foo':'',
                'bar':'1'
            },
            None
        )
        self.assertEqual(expected, parsed)

    def test_parse_user_password_proxy(self):
        parsed = opener.parse('ftp://user:password@ftp.example.org?proxy=ftp.proxy.org')
        expected = ParseResult(
            'ftp',
            'user',
            'password',
            'ftp.example.org',
            {
                'proxy': 'ftp.proxy.org'
            },
            None
        )
        self.assertEqual(expected, parsed)

    def test_parse_user_password_decode(self):
        parsed = opener.parse('ftp://user%40large:password@ftp.example.org')
        expected = ParseResult(
            'ftp',
            'user@large',
            'password',
            'ftp.example.org',
            {},
            None
        )
        self.assertEqual(expected, parsed)

    def test_parse_resource_decode(self):
        parsed = opener.parse('ftp://user%40large:password@ftp.example.org/%7Econnolly')
        expected = ParseResult(
            'ftp',
            'user@large',
            'password',
            'ftp.example.org/~connolly',
            {},
            None
        )
        self.assertEqual(expected, parsed)


class TestRegistry(unittest.TestCase):

    def test_registry_protocols(self):
        # Check registry.protocols list the names of all available entry points

        protocols = [
            entry_point.name
            for entry_point in
            pkg_resources.iter_entry_points('fs.opener')
        ]

        self.assertEqual(
            sorted(protocols),
            sorted(opener.registry.protocols)
        )

    def test_unknown_protocol(self):
        with self.assertRaises(errors.UnsupportedProtocol):
            opener.open_fs('unknown://')

    def test_entry_point_load_error(self):

        entry_point = mock.MagicMock()
        entry_point.load.side_effect = ValueError("some error")

        iter_entry_points = mock.MagicMock(return_value=iter([entry_point]))

        with mock.patch('pkg_resources.iter_entry_points', iter_entry_points):
            with self.assertRaises(errors.EntryPointError) as ctx:
                opener.open_fs('test://')
            self.assertEqual(
                'could not load entry point; some error', str(ctx.exception))

    def test_entry_point_type_error(self):

        entry_point = mock.MagicMock()

        iter_entry_points = mock.MagicMock(return_value=iter([entry_point]))

        with mock.patch('pkg_resources.iter_entry_points', iter_entry_points):
            with self.assertRaises(errors.EntryPointError) as ctx:
                opener.open_fs('test://')
            self.assertEqual(
                'entry point did not return an opener', str(ctx.exception))

    def test_entry_point_create_error(self):

        entry_point = mock.MagicMock()
        entry_point.load = mock.MagicMock(return_value=entry_point)
        entry_point.side_effect = ValueError("some creation error")

        iter_entry_points = mock.MagicMock(return_value=iter([entry_point]))
        builtins = 'builtins' if six.PY3 else '__builtin__'

        with mock.patch('pkg_resources.iter_entry_points', iter_entry_points):
            with mock.patch(builtins+'.issubclass', return_value=True):
                with self.assertRaises(errors.EntryPointError) as ctx:
                    opener.open_fs('test://')
            self.assertEqual(
                'could not instantiate opener; some creation error', str(ctx.exception))


class TestManageFS(unittest.TestCase):

    def test_manage_fs_url(self):
        with opener.manage_fs('mem://') as mem_fs:
            self.assertIsInstance(mem_fs, MemoryFS)
        self.assertTrue(mem_fs.isclosed())

    def test_manage_fs_obj(self):
        mem_fs = MemoryFS()
        with opener.manage_fs(mem_fs) as open_mem_fs:
            self.assertIs(mem_fs, open_mem_fs)
        self.assertFalse(mem_fs.isclosed())

    def test_manage_fs_error(self):
        try:
            with opener.manage_fs('mem://') as mem_fs:
                1/0
        except ZeroDivisionError:
            pass
        self.assertTrue(mem_fs.isclosed())


class TestOpeners(unittest.TestCase):

    def test_repr(self):
        # Check __repr__ works
        for entry_point in pkg_resources.iter_entry_points('fs.opener'):
            _opener = entry_point.load()
            repr(_opener())

    def test_open_osfs(self):
        fs = opener.open_fs("osfs://.")
        self.assertIsInstance(fs, OSFS)

        # test default protocol
        fs = opener.open_fs("./")
        self.assertIsInstance(fs, OSFS)

    def test_open_memfs(self):
        fs = opener.open_fs("mem://")
        self.assertIsInstance(fs, MemoryFS)

    def test_open_zipfs(self):
        fh, zip_name = tempfile.mkstemp()
        os.close(fh)
        try:
            # Test creating zip
            with opener.open_fs('zip://' + zip_name, create=True) as make_zip:
                make_zip.settext('foo.txt', 'foofoo')
            # Test opening zip
            with opener.open_fs('zip://' + zip_name) as zip_fs:
                self.assertEqual(zip_fs.gettext('foo.txt'), 'foofoo')
        finally:
            os.remove(zip_name)

    def test_open_tarfs(self):
        fh, tar_name = tempfile.mkstemp(suffix='.tar.gz')
        os.close(fh)
        try:
            # Test creating tar
            with opener.open_fs('tar://' + tar_name, create=True) as make_tar:
                self.assertEqual(make_tar.compression, 'gz')
                make_tar.settext('foo.txt', 'foofoo')
            # Test opening tar
            with opener.open_fs('tar://' + tar_name) as tar_fs:
                self.assertEqual(tar_fs.gettext('foo.txt'), 'foofoo')
        finally:
            os.remove(tar_name)

    def test_open_fs(self):
        mem_fs = opener.open_fs("mem://")
        mem_fs_2 = opener.open_fs(mem_fs)
        self.assertEqual(mem_fs, mem_fs_2)

    def test_open_userdata(self):
        with self.assertRaises(errors.OpenerError):
            opener.open_fs('userdata://foo:bar:baz:egg')

        app_fs = opener.open_fs(
            'userdata://fstest:willmcgugan:1.0',
            create=True
        )
        self.assertEqual(app_fs.app_dirs.appname, 'fstest')
        self.assertEqual(app_fs.app_dirs.appauthor, 'willmcgugan')
        self.assertEqual(app_fs.app_dirs.version, '1.0')

    def test_open_userdata_no_version(self):
        app_fs = opener.open_fs(
            'userdata://fstest:willmcgugan',
            create=True
        )
        self.assertEqual(app_fs.app_dirs.appname, 'fstest')
        self.assertEqual(app_fs.app_dirs.appauthor, 'willmcgugan')
        self.assertEqual(app_fs.app_dirs.version, None)

    def test_user_data_opener(self):
        user_data_fs = open_fs('userdata://fstest:willmcgugan:1.0')
        self.assertIsInstance(user_data_fs, UserDataFS)
        user_data_fs.makedir('foo', recreate=True)
        user_data_fs.settext('foo/bar.txt', 'baz')
        user_data_fs_foo_dir = open_fs('userdata://fstest:willmcgugan:1.0/foo/')
        self.assertEqual(user_data_fs_foo_dir.gettext('bar.txt'), 'baz')




from __future__ import unicode_literals

import errno
import unittest

from fs.error_tools import convert_os_errors
from fs import errors as fserrors


class TestErrorTools(unittest.TestCase):

    def assert_convert_os_errors(self):

        with self.assertRaises(fserrors.ResourceNotFound):
            with convert_os_errors('foo', 'test'):
                raise OSError(errno.ENOENT)



"""
A collection of filesystems that map to application specific locations
defined by the OS.

These classes abstract away the different requirements for user data
across platforms, which vary in their conventions. They are all
subclasses of :class:`~fs.osfs.OSFS`.

"""
# Thanks to authors of https://pypi.python.org/pypi/appdirs

# see http://technet.microsoft.com/en-us/library/cc766489(WS.10).aspx

from .osfs import OSFS
from ._repr import make_repr
from appdirs import AppDirs

__all__ = ['UserDataFS',
           'UserConfigFS',
           'SiteDataFS',
           'SiteConfigFS',
           'UserCacheFS',
           'UserLogFS']


class _AppFS(OSFS):
    """
    Abstract base class for an app FS.

    """
    app_dir = None

    def __init__(self,
                 appname,
                 author=None,
                 version=None,
                 roaming=False,
                 create=True):
        self.app_dirs = AppDirs(appname, author, version, roaming)
        self.create = create
        super(_AppFS, self).__init__(
            getattr(self.app_dirs, self.app_dir),
            create=create
        )

    def __repr__(self):
        return make_repr(
            self.__class__.__name__,
            self.app_dirs.appname,
            author=(self.app_dirs.appauthor, None),
            version=(self.app_dirs.version, None),
            roaming=(self.app_dirs.roaming, False),
            create=(self.create, True)
        )

    def __str__(self):
        return "<{} '{}'>".format(
            self.__class__.__name__.lower(),
            self.app_dirs.appname
        )

class UserDataFS(_AppFS):
    """
    A filesystem for per-user application data.

    May also be opened with
    ``open_fs('userdata://appname:author:version')``.

    :param str appname: The name of the application.
    :param str author: The name of the author (used on Windows).
    :param str version: Optional version string, if a unique location
        per version of the application is required.
    :param bool roaming: If ``True``, use a *roaming* profile on
        Windows.
    :param bool create: If ``True`` (the default) the directory will
        be created if it does not exist.

    """
    app_dir = 'user_data_dir'


class UserConfigFS(_AppFS):
    """
    A filesystem for per-user config data.

    May also be opened with
    ``open_fs('userconf://appname:author:version')``.

    :param str appname: The name of the application.
    :param str author: The name of the author (used on Windows).
    :param str version: Optional version string, if a unique location
        per version of the application is required.
    :param bool roaming: If ``True``, use a *roaming* profile on
        Windows.
    :param bool create: If ``True`` (the default) the directory will
        be created if it does not exist.

    """
    app_dir = 'user_config_dir'


class UserCacheFS(_AppFS):
    """
    A filesystem for per-user application cache data.

    May also be opened with
    ``open_fs('usercache://appname:author:version')``.

    :param str appname: The name of the application.
    :param str author: The name of the author (used on Windows).
    :param str version: Optional version string, if a unique location
        per version of the application is required.
    :param bool roaming: If ``True``, use a *roaming* profile on
        Windows.
    :param bool create: If ``True`` (the default) the directory will
        be created if it does not exist.

    """
    app_dir = 'user_cache_dir'


class SiteDataFS(_AppFS):
    """
    A filesystem for application site data.

    May also be opened with
    ``open_fs('sitedata://appname:author:version')``.

    :param str appname: The name of the application.
    :param str author: The name of the author (used on Windows).
    :param str version: Optional version string, if a unique location
        per version of the application is required.
    :param bool roaming: If ``True``, use a *roaming* profile on
        Windows.
    :param bool create: If ``True`` (the default) the directory will
        be created if it does not exist.

    """
    app_dir = 'site_data_dir'


class SiteConfigFS(_AppFS):
    """
    A filesystem for application config data.

    May also be opened with
    ``open_fs('siteconf://appname:author:version')``.

    :param str appname: The name of the application.
    :param str author: The name of the author (used on Windows).
    :param str version: Optional version string, if a unique location
        per version of the application is required.
    :param bool roaming: If ``True``, use a *roaming* profile on
        Windows.
    :param bool create: If ``True`` (the default) the directory will
        be created if it does not exist.

    """
    app_dir = 'site_config_dir'


class UserLogFS(_AppFS):
    """
    A filesystem for per-user application log data.

    May also be opened with
    ``open_fs('userlog://appname:author:version')``.

    :param str appname: The name of the application.
    :param str author: The name of the author (used on Windows).
    :param str version: Optional version string, if a unique location
        per version of the application is required.
    :param bool roaming: If ``True``, use a *roaming* profile on
        Windows.
    :param bool create: If ``True`` (the default) the directory will
        be created if it does not exist.

    """
    app_dir = 'user_log_dir'



from __future__ import unicode_literals

import unittest

from fs.mountfs import MountError, MountFS
from fs.memoryfs import MemoryFS
from fs.tempfs import TempFS
from fs.test import FSTestCases


class TestMountFS(FSTestCases, unittest.TestCase):
    """Test OSFS implementation."""

    def make_fs(self):
        fs = MountFS()
        mem_fs = MemoryFS()
        fs.mount('/', mem_fs)
        return fs


class TestMountFS2(FSTestCases, unittest.TestCase):
    """Test OSFS implementation."""

    def make_fs(self):
        fs = MountFS()
        mem_fs = MemoryFS()
        fs.mount('/foo', mem_fs)
        return fs.opendir('foo')


class TestMountFSBehaviours(unittest.TestCase):

    def test_bad_mount(self):
        mount_fs = MountFS()
        with self.assertRaises(TypeError):
            mount_fs.mount('foo', 5)
        with self.assertRaises(TypeError):
            mount_fs.mount('foo', b'bar')

    def test_listdir(self):
        mount_fs = MountFS()
        self.assertEqual(mount_fs.listdir('/'), [])
        m1 = MemoryFS()
        m3 = MemoryFS()
        m4 = TempFS()
        mount_fs.mount('/m1', m1)
        mount_fs.mount('/m2', 'temp://')
        mount_fs.mount('/m3', m3)
        with self.assertRaises(MountError):
            mount_fs.mount('/m3/foo', m4)
        self.assertEqual(
            sorted(mount_fs.listdir('/')),
            ['m1', 'm2', 'm3']
        )
        m3.makedir('foo')
        self.assertEqual(
            sorted(mount_fs.listdir('/m3')),
            ['foo']
        )

    def test_auto_close(self):
        """Test MountFS auto close is working"""
        mount_fs = MountFS()
        m1 = MemoryFS()
        m2 = MemoryFS()
        mount_fs.mount('/m1', m1)
        mount_fs.mount('/m2', m2)
        self.assertFalse(m1.isclosed())
        self.assertFalse(m2.isclosed())
        mount_fs.close()
        self.assertTrue(m1.isclosed())
        self.assertTrue(m2.isclosed())

    def test_no_auto_close(self):
        """Test MountFS auto close can be disabled"""
        mount_fs = MountFS(auto_close=False)
        m1 = MemoryFS()
        m2 = MemoryFS()
        mount_fs.mount('/m1', m1)
        mount_fs.mount('/m2', m2)
        self.assertFalse(m1.isclosed())
        self.assertFalse(m2.isclosed())
        mount_fs.close()
        self.assertFalse(m1.isclosed())
        self.assertFalse(m2.isclosed())

    def test_empty(self):
        """Test MountFS with nothing mounted."""
        mount_fs = MountFS()
        self.assertEqual(mount_fs.listdir('/'), [])

    def test_mount_self(self):
        mount_fs = MountFS()
        with self.assertRaises(ValueError):
            mount_fs.mount('/', mount_fs)

    def test_desc(self):
        mount_fs = MountFS()
        mount_fs.desc('/')



from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from six import text_type

from . import errors
from .base import FS
from .memoryfs import MemoryFS
from .path import abspath
from .path import forcedir
from .path import normpath
from .mode import validate_open_mode
from .mode import validate_openbin_mode


class MountError(Exception):
    """Thrown when mounts conflict."""


class MountFS(FS):
    """
    A virtual filesystem that maps directories on to other file-systems.

    :param auto_close: If True, the child filesystems will be closed
        when the ``MountFS`` is closed.
    :type auto_close: bool

    """

    _meta = {
        "virtual": True,
        "read_only": False,
        'unicode_paths': True,
        "case_insensitive": False,
        "invalid_path_chars": "\0",
    }

    def __init__(self, auto_close=True):
        super(MountFS, self).__init__()
        self.auto_close = auto_close
        self.default_fs = MemoryFS()
        self.mounts = []

    def __repr__(self):
        return "MountFS(auto_close={!r})".format(self.auto_close)

    def __str__(self):
        return "<mountfs>"

    def _delegate(self, path):
        """
        Get a tuple of (fs, path) for a mounted filesystem, or (None,
        None) if no filesystem is mounted on the given path.

        """
        _path = forcedir(abspath(normpath(path)))
        is_mounted = _path.startswith

        for mount_path, fs in self.mounts:
            if is_mounted(mount_path):
                return fs, _path[len(mount_path):].rstrip('/')

        return self.default_fs, path

    def mount(self, path, fs):
        """
        Mounts a host FS object on a given path.

        :param path: A path within the MountFS.
        :type path: str
        :param fs: A filesystem object or FS URL to mount.
        :type fs: :class:`~fs.base.FS`

        """
        if isinstance(fs, text_type):
            from .opener import open_fs
            fs = open_fs(fs)

        if not isinstance(fs, FS):
            raise TypeError(
                'fs argument must be an FS object or a FS URL'
            )

        if fs is self:
            raise ValueError('Unable to mount self')
        _path = forcedir(abspath(normpath(path)))

        for mount_path, _ in self.mounts:
            if _path.startswith(mount_path):
                raise MountError(
                    "mount point overlaps existing mount"
                )

        self.mounts.append((_path, fs))
        self.default_fs.makedirs(_path, recreate=True)

    def close(self):
        # Explicitly closes children if requested
        if self.auto_close:
            for _path, fs in self.mounts:
                fs.close()
            del self.mounts[:]
        self.default_fs.close()
        super(MountFS, self).close()

    def desc(self, path):
        if not self.exists(path):
            raise errors.ResourceNotFound(path)
        fs, delegate_path = self._delegate(path)
        if fs is self.default_fs:
            fs = self
        return "{path} on {fs}".format(fs=fs, path=delegate_path)

    def getinfo(self, path, namespaces=None):
        self.check()
        fs, _path = self._delegate(path)
        return fs.getinfo(_path, namespaces=namespaces)

    def listdir(self, path):
        self.check()
        fs, _path = self._delegate(path)
        return fs.listdir(_path)

    def makedir(self, path, permissions=None, recreate=False):
        self.check()
        fs, _path = self._delegate(path)
        return fs.makedir(
            _path, permissions=permissions, recreate=recreate)

    def openbin(self, path, mode='r', buffering=-1, **kwargs):
        validate_openbin_mode(mode)
        self.check()
        fs, _path = self._delegate(path)
        return fs.openbin(_path, mode=mode, buffering=-1, **kwargs)

    def remove(self, path):
        self.check()
        fs, _path = self._delegate(path)
        return fs.remove(_path)

    def removedir(self, path):
        self.check()
        path = normpath(path)
        if path in ('', '/'):
            raise errors.RemoveRootError(path)
        fs, _path = self._delegate(path)
        return fs.removedir(_path)

    def getbytes(self, path):
        self.check()
        fs, _path = self._delegate(path)
        return fs.getbytes(_path)

    def gettext(self, path, encoding=None, errors=None, newline=None):
        self.check()
        fs, _path = self._delegate(path)
        return fs.gettext(
            _path,
            encoding=encoding,
            errors=errors,
            newline=newline
        )

    def getsize(self, path):
        self.check()
        fs, _path = self._delegate(path)
        return fs.getsize(_path)

    def getsyspath(self, path):
        self.check()
        fs, _path = self._delegate(path)
        return fs.getsyspath(_path)

    def gettype(self, path):
        self.check()
        fs, _path = self._delegate(path)
        return fs.gettype(_path)

    def geturl(self, path, purpose='download'):
        self.check()
        fs, _path = self._delegate(path)
        return fs.geturl(_path, purpose=purpose)

    def hasurl(self, path, purpose='download'):
        self.check()
        fs, _path = self._delegate(path)
        return fs.hasurl(_path, purpose=purpose)

    def isdir(self, path):
        self.check()
        fs, _path = self._delegate(path)
        return fs.isdir(_path)

    def isfile(self, path):
        self.check()
        fs, _path = self._delegate(path)
        return fs.isfile(_path)

    def scandir(self, path, namespaces=None, page=None):
        self.check()
        fs, _path = self._delegate(path)
        return fs.scandir(_path, namespaces=namespaces, page=page)

    def setinfo(self, path, info):
        self.check()
        fs, _path = self._delegate(path)
        return fs.setinfo(_path, info)

    def validatepath(self, path):
        self.check()
        fs, _path = self._delegate(path)
        return fs.validatepath(_path)

    def open(self,
             path,
             mode='r',
             buffering=-1,
             encoding=None,
             errors=None,
             newline='',
             **options):
        validate_open_mode(mode)
        self.check()
        fs, _path = self._delegate(path)
        return fs.open(
            _path,
            mode=mode,
            buffering=buffering,
            encoding=encoding,
            errors=errors,
            newline=newline,
            **options
        )

    def setbinfile(self, path, file):
        self.check()
        fs, _path = self._delegate(path)
        return fs.setbinfile(_path, file)

    def setbytes(self, path, contents):
        self.check()
        fs, _path = self._delegate(path)
        return fs.setbytes(_path, contents)

    def settext(self,
                path,
                contents,
                encoding='utf-8',
                errors=None,
                newline=''):
        fs, _path = self._delegate(path)
        return fs.settext(
            _path,
            contents,
            encoding=encoding,
            errors=errors,
            newline=newline
        )



"""
fs.filesize
===========

Functions for reporting filesizes

"""

from __future__ import division
from __future__ import unicode_literals

__all__ = ['traditional', 'decimal', 'binary']



def _to_str(size, suffixes, base):
    try:
        size = int(size)
    except ValueError:
        raise TypeError(
            "filesize requires a numeric value, not {!r}".format(size)
        )
    if size == 1:
        return '1 byte'
    elif size < base:
        return '{:,} bytes'.format(size)

    for i, suffix in enumerate(suffixes, 2):
        unit = base ** i
        if size < unit:
            break
    return "{:,.1f} {}".format((base * size / unit), suffix)


def traditional(size):
    """
    Convert a filesize in to a string representation with traditional
    (base 2) units and JDEC prefixes.

    :param int size: A file size.
    :returns: A string containing a abbreviated file size and units.
    :rtype str:

    """
    return _to_str(
        size,
        ('KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB'),
        1024
    )


def binary(size):
    """
    Convert a filesize in to a string representation with binary units
    and SI binary prefixes.

    :param int size: A file size.
    :param bool si: True to use SI prefixes, False to use JDEC prefixes.
    :returns: A string containing a abbreviated file size and units.
    :rtype str:

    """
    return _to_str(
        size,
        ('KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB', 'YiB'),
        1024
    )


def decimal(size):
    """
    Convert a filesize in to a string representation with decimal
    units and SI decimal prefixes.

    :param int size: A file size.
    :returns: A string containing a abbreviated file size and units.
    :rtype str:
    """
    return _to_str(
        size,
        ('kB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB'),
        1000
    )



from __future__ import unicode_literals
from __future__ import print_function

import unittest

from six import text_type

from fs.permissions import make_mode, Permissions


class TestPermissions(unittest.TestCase):

    def test_make_mode(self):
        self.assertEqual(make_mode(None), 0o777)
        self.assertEqual(make_mode(0o755), 0o755)
        self.assertEqual(make_mode(['u_r', 'u_w', 'u_x']), 0o700)
        self.assertEqual(make_mode(Permissions(user='rwx')), 0o700)


    def test_parse(self):
        self.assertEqual(Permissions.parse('---------').mode, 0)
        self.assertEqual(Permissions.parse('rwxrw-r--').mode, 0o764)

    def test_create(self):
        self.assertEqual(Permissions.create(None).mode, 0o777)
        self.assertEqual(Permissions.create(0o755).mode, 0o755)
        self.assertEqual(Permissions.create(['u_r', 'u_w', 'u_x']).mode, 0o700)
        self.assertEqual(
            Permissions.create(Permissions(user='rwx')).mode,
            0o700
        )
        with self.assertRaises(ValueError):
            Permissions.create('foo')

    def test_constructor(self):
        p = Permissions(names=['foo', 'bar'])
        self.assertIn('foo', p)
        self.assertIn('bar', p)
        self.assertNotIn('baz', p)

        p = Permissions(user='r', group='w', other='x')
        self.assertIn('u_r', p)
        self.assertIn('g_w', p)
        self.assertIn('o_x', p)
        self.assertNotIn('sticky', p)
        self.assertNotIn('setuid', p)
        self.assertNotIn('setguid', p)

        p = Permissions(user='rwx', group='rwx', other='rwx',
                        sticky=True, setuid=True, setguid=True)
        self.assertIn('sticky', p)
        self.assertIn('setuid', p)
        self.assertIn('setguid', p)

        p = Permissions(mode=0o421)
        self.assertIn('u_r', p)
        self.assertIn('g_w', p)
        self.assertIn('o_x', p)
        self.assertNotIn('u_w', p)
        self.assertNotIn('g_x', p)
        self.assertNotIn('o_r', p)
        self.assertNotIn('sticky', p)
        self.assertNotIn('setuid', p)
        self.assertNotIn('setguid', p)

    def test_properties(self):
        p = Permissions()
        self.assertFalse(p.u_r)
        self.assertNotIn('u_r', p)
        p.u_r = True
        self.assertIn('u_r', p)
        self.assertTrue(p.u_r)
        p.u_r = False
        self.assertNotIn('u_r', p)
        self.assertFalse(p.u_r)

        self.assertFalse(p.u_w)
        p.add('u_w')
        self.assertTrue(p.u_w)
        p.remove('u_w')
        self.assertFalse(p.u_w)

    def test_repr(self):
        self.assertEqual(
            repr(Permissions()),
            "Permissions(user='', group='', other='')",
        )
        self.assertEqual(
            repr(Permissions(names=['foo'])),
            "Permissions(names=['foo'])"
        )
        repr(Permissions(user='rwx', group='rw', other='r'))
        repr(Permissions(user='rwx', group='rw', other='r', sticky=True))
        repr(Permissions(user='rwx', group='rw', other='r', setuid=True))
        repr(Permissions(user='rwx', group='rw', other='r', setguid=True))

    def test_as_str(self):
        p = Permissions(user='rwx', group='rwx', other='rwx')
        self.assertEqual(p.as_str(), 'rwxrwxrwx')
        self.assertEqual(str(p), 'rwxrwxrwx')
        p = Permissions(mode=0o777, setuid=True, setguid=True, sticky=True)
        self.assertEqual(p.as_str(), 'rwsrwsrwt')

    def test_mode(self):
        p = Permissions(user='rwx', group='rw', other='')
        self.assertEqual(p.mode, 0o760)

    def test_serialize(self):
        p = Permissions(names=['foo'])
        self.assertEqual(p.dump(), ['foo'])
        pp = Permissions.load(['foo'])
        self.assertIn('foo', pp)

    def test_iter(self):
        iter_p = iter(Permissions(names=['foo']))
        self.assertEqual(list(iter_p), ['foo'])

    def test_equality(self):
        self.assertEqual(Permissions(mode=0o700), Permissions(user='rwx'))
        self.assertNotEqual(Permissions(mode=0o500), Permissions(user='rwx'))

        self.assertEqual(Permissions(mode=0o700), ['u_r', 'u_w', 'u_x'])

    def test_copy(self):
        p = Permissions(mode=0o700)
        p_copy = p.copy()
        self.assertIsNot(p, p_copy)
        self.assertEqual(p, p_copy)

    def test_check(self):
        p = Permissions(user='rwx')
        self.assertTrue(p.check('u_r'))
        self.assertTrue(p.check('u_r', 'u_w'))
        self.assertTrue(p.check('u_r', 'u_w', 'u_x'))
        self.assertFalse(p.check('u_r', 'g_w'))
        self.assertFalse(p.check('g_r', 'g_w'))
        self.assertFalse(p.check('foo'))

    def test_mode_set(self):
        p = Permissions(user='r')
        self.assertEqual(text_type(p), 'r--------')
        p.mode = 0o700
        self.assertEqual(text_type(p), 'rwx------')



"""

Functions for moving files between filesystems.

"""

from __future__ import print_function
from __future__ import unicode_literals

from .copy import copy_dir
from .copy import copy_file
from .opener import manage_fs


def move_fs(src_fs, dst_fs):
    """
    Move the contents of a filesystem to another filesystem.

    :param src_fs: Source filesystem.
    :type src_fs: FS URL or instance
    :param dst_fs: Destination filesystem.
    :type dst_fs: FS URL or instance

    """
    move_dir(src_fs, '/', dst_fs, '/')


def move_file(src_fs, src_path, dst_fs, dst_path):
    """
    Move a file from one filesystem to another.

    :param src_fs: Source filesystem.
    :type src_fs: FS URL or instance
    :param src_path: Path to a file on ``src_fs``.
    :type src_path: str
    :param dst_fs: Destination filesystem.
    :type dst_fs: FS URL or instance
    :param dst_fs: Path to a file on ``dst_fs``.
    :type dst_fs: str

    """
    with manage_fs(src_fs) as src_fs:
        with manage_fs(dst_fs, create=True) as dst_fs:
            if src_fs is dst_fs:
                # Same filesystem, may be optimized
                src_fs.move(src_path, dst_path, overwrite=True)
            else:
                # Standard copy and delete
                with src_fs.lock(), dst_fs.lock():
                    copy_file(src_fs, src_path, dst_fs, dst_path)
                    src_fs.remove(src_path)


def move_dir(src_fs, src_path, dst_fs, dst_path):
    """
    Move a directory from one filesystem to another.

    :param src_fs: Source filesystem.
    :type src_fs: FS URL or instance
    :param src_path: A path to a directory on ``src_fs``.
    :type src_path: str
    :param dst_fs: Destination filesystem.
    :type dst_fs: FS URL or instance
    :param str dst_path: A path to a directory on ``dst_fs``.

    """
    with manage_fs(src_fs) as src_fs:
        with manage_fs(dst_fs, create=True) as dst_fs:
            with src_fs.lock(), dst_fs.lock():
                dst_fs.makedir(dst_path, recreate=True)
                copy_dir(
                    src_fs,
                    src_path,
                    dst_fs,
                    dst_path
                )
                src_fs.removetree(src_path)



# coding: utf-8
"""Defines the TempOpener."""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from .base import Opener

class TempOpener(Opener):
    protocols = ['temp']

    def open_fs(self, fs_url, parse_result, writeable, create, cwd):
        from ..tempfs import TempFS
        temp_fs = TempFS(identifier=parse_result.resource)
        return temp_fs



"""Test (abstract) base FS class."""

from __future__ import unicode_literals

import unittest

try:
    import mock
except ImportError:
    from unittest import mock


from fs.base import FS
from fs import errors


class TestFS(FS):

    def getinfo(self, path, namespaces=None):
        pass

    def listdir(self, path):
        pass

    def makedir(self, path, permissions=None, recreate=False):
        pass

    def openbin(self, path, mode="r", buffering=-1, **options):
        pass

    def remove(self, path):
        pass

    def removedir(self, path):
        pass

    def setinfo(self, path, info):
        pass


class TestBase(unittest.TestCase):
    def setUp(self):
        self.fs = TestFS()

    def test_validatepath(self):
        """Test validatepath method."""
        with self.assertRaises(TypeError):
            self.fs.validatepath(b'bytes')

        self.fs._meta['invalid_path_chars'] = 'Z'
        with self.assertRaises(errors.InvalidCharsInPath):
            self.fs.validatepath('Time for some ZZZs')

        self.fs.validatepath('fine')
        self.fs.validatepath('good.fine')

        self.fs._meta['invalid_path_chars'] = ''
        self.fs.validatepath('Time for some ZZZs')

        def mock_getsyspath(path):
            return path
        self.fs.getsyspath = mock_getsyspath

        self.fs._meta['max_sys_path_length'] = 10

        self.fs.validatepath('0123456789')
        self.fs.validatepath('012345678')
        self.fs.validatepath('01234567')

        with self.assertRaises(errors.InvalidPath):
            self.fs.validatepath('0123456789A')



"""Match wildcard filenames."""
# Adapted from https://hg.python.org/cpython/file/2.7/Lib/fnmatch.py

from __future__ import unicode_literals

import re
from functools import partial

from .lrucache import LRUCache

_MAXCACHE = 1000
_PATTERN_CACHE = LRUCache(_MAXCACHE)


def match(pattern, name):
    """
    Test whether ``name`` matches ``pattern``.

    :param str pattern: A wildcard pattern. e.g. ``"*.py"``
    :param str name: A filename
    :rtype: bool

    """
    try:
        re_pat = _PATTERN_CACHE[(pattern, True)]
    except KeyError:
        res = _translate(pattern)
        _PATTERN_CACHE[(pattern, True)] = re_pat = re.compile(res)
    return re_pat.match(name) is not None


def imatch(pattern, name):
    """
    Test whether ``name`` matches ``pattern``, ignoring
    case differences.

    :param str pattern: A wildcard pattern. e.g. ``"*.py"``
    :param str name: A filename
    :rtype: bool

    """
    try:
        re_pat = _PATTERN_CACHE[(pattern, False)]
    except KeyError:
        res = _translate(pattern, case_sensitive=False)
        _PATTERN_CACHE[(pattern, False)] = re_pat =\
            re.compile(res, re.IGNORECASE)
    return re_pat.match(name) is not None


def match_any(patterns, name):
    """
    Test if a name matches at least one of a list of patterns. Will
    return ``True`` if ``patterns`` is an empty list.

    :param list patterns: A list of wildcard pattern. e.g. ``["*.py",
        "*.pyc"]``
    :param str name: A filename.
    :rtype: bool

    """
    if not patterns:
        return True
    return any(match(pattern, name) for pattern in patterns)


def imatch_any(patterns, name):
    """
    Test if a name matches at least one of a list of patterns, ignoring
    case differences. Will return ``True`` if ``patterns`` is an empty
    list.

    :param list patterns: A list of wildcard pattern. e.g. ``["*.py",
        "*.pyc"]``
    :param str name: A filename.
    :rtype: bool

    """
    if not patterns:
        return True
    return any(imatch(pattern, name) for pattern in patterns)


def get_matcher(patterns, case_sensitive):
    """
    Get a callable that checks a list of names matches the given
    wildcard patterns.

    :param list patterns: A list of wildcard pattern. e.g. ``["*.py",
        "*.pyc"]``
    :param bool case_sensitive: If True, then the callable will be case
        sensitive, otherwise it will be case insensitive.
    :rtype: callable

    Here's an example::

    >>> import wildcard
    >>> is_python = wildcard.get_macher(['*.py'])
    >>> is_python('__init__.py')
    >>> True
    >>> is_python('foo.txt')
    >>> False

    """
    if not patterns:
        return lambda name: True
    if case_sensitive:
        return partial(match_any, patterns)
    else:
        return partial(imatch_any, patterns)


def _translate(pattern, case_sensitive=True):
    """
    Translate a shell PATTERN to a regular expression.

    There is no way to quote meta-characters.

    """
    if not case_sensitive:
        pattern = pattern.lower()
    i, n = 0, len(pattern)
    res = ''
    while i < n:
        c = pattern[i]
        i = i + 1
        if c == '*':
            res = res + '.*'
        elif c == '?':
            res = res + '.'
        elif c == '[':
            j = i
            if j < n and pattern[j] == '!':
                j = j + 1
            if j < n and pattern[j] == ']':
                j = j + 1
            while j < n and pattern[j] != ']':
                j = j + 1
            if j >= n:
                res = res + '\\['
            else:
                stuff = pattern[i:j].replace('\\', '\\\\')
                i = j + 1
                if stuff[0] == '!':
                    stuff = '^' + stuff[1:]
                elif stuff[0] == '^':
                    stuff = '\\' + stuff
                res = '%s[%s]' % (res, stuff)
        else:
            res = res + re.escape(c)
    return res + '\Z(?ms)'



from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import unicodedata
import datetime
import re
import time

from pytz import UTC

from .enums import ResourceType
from .permissions import Permissions


epoch_dt = datetime.datetime.fromtimestamp(0, UTC)


re_linux = re.compile(
    """
    ^
    ([ldrwx-]{10})
    \s+?
    (\d+)
    \s+?
    (\w+)
    \s+?
    (\w+)
    \s+?
    (\d+)
    \s+?
    (\w{3}\s+\d{1,2}\s+[\w:]+)
    \s+
    (.*?)
    $
    """,
    re.VERBOSE
)


def get_decoders():
    decoders = [
        (re_linux, decode_linux),
    ]
    return decoders


def parse(lines):
    info = []
    for line in lines:
        if not line.strip():
            continue
        raw_info = parse_line(line)
        if raw_info is not None:
            info.append(raw_info)
    return info


def parse_line(line):
    for line_re, decode_callable in get_decoders():
        match = line_re.match(line)
        if match is not None:
            return decode_callable(line, match)
    return None


def _parse_time(t):
    t = ' '.join(token.strip() for token in t.lower().split(' '))
    try:
        try:
            _t = time.strptime(t, '%b %d %Y')
        except ValueError:
            _t = time.strptime(t, '%b %d %H:%M')
    except ValueError:
        # Unknown time format
        return None

    year = _t.tm_year if _t.tm_year != 1900 else time.localtime().tm_year
    month = _t.tm_mon
    day = _t.tm_mday
    hour = _t.tm_hour
    minutes = _t.tm_min
    dt = datetime.datetime(
        year, month, day,
        hour, minutes,
        tzinfo=UTC
    )

    epoch_time = (dt - epoch_dt).total_seconds()
    return epoch_time


def decode_linux(line, match):
    perms, links, uid, gid, size, mtime, name = match.groups()
    is_link = perms.startswith('l')
    is_dir = perms.startswith('d') or is_link
    if is_link:
        name, _, _link_name = name.partition('->')
        name = name.strip()
        _link_name = _link_name.strip()
    permissions = Permissions.parse(perms[1:])

    mtime_epoch = _parse_time(mtime)

    name = unicodedata.normalize('NFC', name)

    raw_info = {
        "basic": {
            "name": name,
            "is_dir": is_dir
        },
        "details": {
            "size": int(size),
            "type": int(
                ResourceType.directory
                if is_dir else
                ResourceType.file
            )
        },
        "access": {
            "permissions": permissions.dump()
        },
        "ftp": {
            "ls": line
        }
    }
    access = raw_info['access']
    details = raw_info['details']
    if mtime_epoch is not None:
        details['modified'] = mtime_epoch

    access['user'] = uid
    access['group'] = gid

    return raw_info



"""
Tools to generate __repr__ strings

"""

from __future__ import unicode_literals

def make_repr(class_name, *args, **kwargs):
    """
    Generate a repr string.

    Positional arguments should be the positional arguments used to
    construct the class. Keyword arguments should consist of tuples of
    the attribute value and default. If the value is the default, then
    it won't be rendered in the output.

    Here's an example::

        def __repr__(self):
            return make_repr('MyClass', 'foo', name=(self.name, None))

    The output of this would be something line ``MyClass('foo',
    name='Will')``.

    """
    arguments = [repr(arg) for arg in args]
    arguments.extend([
        "{}={!r}".format(name, value)
        for name, (value, default) in sorted(kwargs.items())
        if value != default
    ])
    return "{}({})".format(class_name, ', '.join(arguments))



from __future__ import unicode_literals

import io
import unittest

import six

from fs import iotools
from fs import tempfs

from fs.test import UNICODE_TEXT


class TestIOTools(unittest.TestCase):

    def setUp(self):
        self.fs = tempfs.TempFS('iotoolstest')

    def tearDown(self):
        self.fs.close()
        del self.fs

    def test_make_stream(self):
        """Test make_stream"""

        self.fs.setbytes('foo.bin', b'foofoo')

        with self.fs.openbin('foo.bin') as f:
            data = f.read()
            self.assert_(isinstance(data, bytes))

        with self.fs.openbin('text.txt', 'wb') as f:
            f.write(UNICODE_TEXT.encode('utf-8'))

        with self.fs.openbin('text.txt') as f:
            with iotools.make_stream("text.txt", f, 'rt') as f2:
                repr(f2)
                text = f2.read()
                self.assertIsInstance(text, six.text_type)

    def test_readinto(self):

        self.fs.setbytes('bytes.bin', b'foofoobarbar')

        with self.fs.openbin('bytes.bin') as bin_file:
            with iotools.make_stream('bytes.bin', bin_file, 'rb') as f:
                data = bytearray(3)
                bytes_read = f.readinto(data)
                self.assertEqual(bytes_read, 3)
                self.assertEqual(bytes(data), b'foo')
                self.assertEqual(f.readline(1), b'f')

        def no_readinto(size):
            raise AttributeError

        with self.fs.openbin('bytes.bin') as bin_file:
            bin_file.readinto = no_readinto
            with iotools.make_stream('bytes.bin', bin_file, 'rb') as f:
                data = bytearray(3)
                bytes_read = f.readinto(data)
                self.assertEqual(bytes_read, 3)
                self.assertEqual(bytes(data), b'foo')
                self.assertEqual(f.readline(1), b'f')


    def test_readinto1(self):

        self.fs.setbytes('bytes.bin', b'foofoobarbar')

        with self.fs.openbin('bytes.bin') as bin_file:
            with iotools.make_stream('bytes.bin', bin_file, 'rb') as f:
                data = bytearray(3)
                bytes_read = f.readinto1(data)
                self.assertEqual(bytes_read, 3)
                self.assertEqual(bytes(data), b'foo')
                self.assertEqual(f.readline(1), b'f')

        def no_readinto(size):
            raise AttributeError

        with self.fs.openbin('bytes.bin') as bin_file:
            bin_file.readinto = no_readinto
            with iotools.make_stream('bytes.bin', bin_file, 'rb') as f:
                data = bytearray(3)
                bytes_read = f.readinto1(data)
                self.assertEqual(bytes_read, 3)
                self.assertEqual(bytes(data), b'foo')
                self.assertEqual(f.readline(1), b'f')

    def test_isatty(self):
        with self.fs.openbin('text.txt', 'wb') as f:
            with iotools.make_stream("text.txt", f, 'wb') as f1:
                self.assertFalse(f1.isatty())

    def test_readlines(self):
        self.fs.setbytes('foo', b'barbar\nline1\nline2')
        with self.fs.open('foo', 'rb') as f:
            f = iotools.make_stream('foo', f, 'rb')
            self.assertEqual(
                list(f),
                [b'barbar\n', b'line1\n', b'line2']
            )
        with self.fs.open('foo', 'rt') as f:
            f = iotools.make_stream('foo', f, 'rb')
            self.assertEqual(
                f.readlines(),
                ['barbar\n', 'line1\n', 'line2']
            )

    def test_readall(self):
        self.fs.setbytes('foo', b'foobar')
        with self.fs.open('foo', 'rt') as f:
            self.assertEqual(f.read(), 'foobar')

    def test_writelines(self):
        with self.fs.open('foo', 'wb') as f:
            f = iotools.make_stream('foo', f, 'rb')
            f.writelines([b'foo', b'bar', b'baz'])
        self.assertEqual(self.fs.getbytes('foo'), b'foobarbaz')

    def test_seekable(self):

        f = io.BytesIO(b'HelloWorld')
        raw_wrapper = iotools.RawWrapper(f)
        self.assertTrue(raw_wrapper.seekable())

        def no_seekable():
            raise AttributeError('seekable')

        f.seekable = no_seekable

        def seek(pos, whence):
            raise IOError('no seek')

        raw_wrapper.seek = seek

        self.assertFalse(raw_wrapper.seekable())

    def test_line_iterator(self):
        f = io.BytesIO(b'Hello\nWorld\n\nfoo')
        self.assertEqual(
            list(iotools.line_iterator(f)),
            [b'Hello\n', b'World\n', b'\n', b'foo']
        )

        f = io.BytesIO(b'Hello\nWorld\n\nfoo')
        self.assertEqual(
            list(iotools.line_iterator(f, 10)),
            [b'Hello\n', b'Worl']
        )

    def test_make_stream_writer(self):
        f = io.BytesIO()
        s = iotools.make_stream('foo', f, 'wb', buffering=1)
        self.assertIsInstance(s, io.BufferedWriter)
        s.write(b'Hello')
        self.assertEqual(f.getvalue(), b'Hello')

    def test_make_stream_reader(self):
        f = io.BytesIO(b'Hello')
        s = iotools.make_stream('foo', f, 'rb', buffering=1)
        self.assertIsInstance(s, io.BufferedReader)
        self.assertEqual(s.read(), b'Hello')

    def test_make_stream_reader_writer(self):
        f = io.BytesIO(b'Hello')
        s = iotools.make_stream('foo', f, '+b', buffering=1)
        self.assertIsInstance(s, io.BufferedRandom)
        self.assertEqual(s.read(), b'Hello')
        s.write(b' World')
        self.assertEqual(f.getvalue(), b'Hello World')





"""
fs.opener.parse
===============

Parses FS URLs in to their constituent parts.

"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import collections
import re

from six.moves.urllib.parse import parse_qs, unquote

from .errors import ParseError


_ParseResult = collections.namedtuple(
    'ParseResult',
    [
        'protocol',
        'username',
        'password',
        'resource',
        'params',
        'path'
    ]
)

class ParseResult(_ParseResult):
    """A named tuple containing fields of a parsed FS URL.

    * ``protocol`` The protocol part of the url, e.g. ``osfs`` or
      ``ftp``.
    * ``username`` A username, or ``None`` .
    * ``password`` An password, or ``None``.
    * ``resource`` A *resource*, typically a domain and path, e.g.
      ``ftp.example.org/dir``
    * ``params`` An dictionary of parameters extracted from the query
      string.
    * ``path`` An optional path within the filesystem.

    """


_RE_FS_URL = re.compile(r'''
^
(.*?)
:\/\/

(?:
(?:(.*?)@(.*?))
|(.*?)
)

(?:
!(.*?)$
)*$
''', re.VERBOSE)


def parse_fs_url(fs_url):
    """
    Parse a Filesystem URL and return a
    :class:`~fs.opener.parse.ParseResult`, or raise
    :class:`~fs.errors.ParseError` (subclass of ValueError) if the FS URL is
    not value.

    :param str fs_url: A filesystem URL
    :rtype: :class:`~fs.opener.parse.ParseResult`

    """

    match = _RE_FS_URL.match(fs_url)
    if match is None:
        raise ParseError('{!r} is not a fs2 url'.format(fs_url))

    fs_name, credentials, url1, url2, path = match.groups()
    if credentials:
        username, _, password = credentials.partition(':')
        username = unquote(username)
        password = unquote(password)
        url = url1
    else:
        username = None
        password = None
        url = url2
    url, has_qs, _params = url.partition('?')
    resource = unquote(url)
    if has_qs:
        params = parse_qs(_params, keep_blank_values=True)
        params = {k:v[0] for k, v in params.items()}
    else:
        params = {}
    return ParseResult(
        fs_name,
        username,
        password,
        resource,
        params,
        path
    )



from __future__ import unicode_literals

import unittest

from fs import memoryfs
from fs.test import FSTestCases


class TestMemoryFS(FSTestCases, unittest.TestCase):
    """Test OSFS implementation."""

    def make_fs(self):
        return memoryfs.MemoryFS()



# coding: utf-8
"""Defines the TarOpener."""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from .base import Opener

class TarOpener(Opener):
    protocols = ['tar']

    def open_fs(self, fs_url, parse_result, writeable, create, cwd):
        from ..tarfs import TarFS
        tar_fs = TarFS(
            parse_result.resource,
            write=create
        )
        return tar_fs




from __future__ import unicode_literals

import unittest

import fs.move
from fs import open_fs


class TestMove(unittest.TestCase):

    def test_move_fs(self):
        src_fs = open_fs('mem://')
        src_fs.makedirs('foo/bar')
        src_fs.touch('test.txt')
        src_fs.touch('foo/bar/baz.txt')

        dst_fs = open_fs('mem://')
        fs.move.move_fs(src_fs, dst_fs)

        self.assertTrue(dst_fs.isdir('foo/bar'))
        self.assertTrue(dst_fs.isfile('test.txt'))
        self.assertTrue(src_fs.isempty('/'))

    def test_copy_dir(self):
        src_fs = open_fs('mem://')
        src_fs.makedirs('foo/bar')
        src_fs.touch('test.txt')
        src_fs.touch('foo/bar/baz.txt')

        dst_fs = open_fs('mem://')
        fs.move.move_dir(src_fs, '/foo', dst_fs, '/')

        self.assertTrue(dst_fs.isdir('bar'))
        self.assertTrue(dst_fs.isfile('bar/baz.txt'))
        self.assertFalse(src_fs.exists('foo'))


# coding: utf-8
"""Defines the MemOpener."""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from .base import Opener

class MemOpener(Opener):
    protocols = ['mem']

    def open_fs(self, fs_url, parse_result, writeable, create, cwd):
        from ..memoryfs import MemoryFS
        mem_fs = MemoryFS()
        return mem_fs



from __future__ import print_function
from __future__ import unicode_literals

import shutil
import tempfile

import six

from . import errors
from .osfs import OSFS


@six.python_2_unicode_compatible
class TempFS(OSFS):
    """
    Create a temporary filesystem.

    :param str identifier: A string to distinguish the directory within
        the OS temp location, used as part of the directory name.
    :param str temp_dir: An OS path to your temp directory (leave as
        ``None`` to auto-detect)
    :param bool auto_clean: If True, the directory contents will be
        wiped on close.
    :param bool ignore_clean_errors: If True, any errors in the clean
        process will be raised. If False, they will be suppressed.

    """

    def __init__(self,
                 identifier=None,
                 temp_dir=None,
                 auto_clean=True,
                 ignore_clean_errors=True):
        self.identifier = identifier
        self._auto_clean = auto_clean
        self._ignore_clean_errors = ignore_clean_errors
        self._cleaned = False

        self.identifier = (identifier or '__tempfs__').replace('/', '-')

        self._temp_dir = tempfile.mkdtemp(
            identifier or "fsTempFS",
            dir=temp_dir
        )
        super(TempFS, self).__init__(self._temp_dir)

    def __repr__(self):
        return "TempFS()"

    def __str__(self):
        return "<tempfs '{}'>".format(self._temp_dir)

    def close(self):
        if self._auto_clean:
            self.clean()
        super(TempFS, self).close()

    def clean(self):
        """Clean (delete) temporary files created by this filesystem."""
        if self._cleaned:
            return

        try:
            shutil.rmtree(self._temp_dir)
        except Exception as error:
            if not self._ignore_clean_errors:
                raise errors.OperationFailed(
                    msg="failed to remove temporary directory",
                    exc=error
                )
        self._cleaned = True



from __future__ import unicode_literals

import os
import tempfile
import unittest
import zipfile

from fs import zipfs
from fs.compress import write_zip
from fs.opener import open_fs
from fs.test import FSTestCases

from .test_archives import ArchiveTestCases


class TestWriteZipFS(FSTestCases, unittest.TestCase):
    """
    Test ZIPFS implementation.

    When writing, a ZipFS is essentially a TempFS.

    """

    def make_fs(self):
        _zip_file = tempfile.TemporaryFile()
        fs = zipfs.ZipFS(_zip_file, write=True)
        fs._zip_file = _zip_file
        return fs

    def destroy_fs(self, fs):
        fs.close()
        del fs._zip_file


class TestReadZipFS(ArchiveTestCases, unittest.TestCase):
    """
    Test Reading zip files.

    """
    def compress(self, fs):
        fh, self._temp_path = tempfile.mkstemp()
        os.close(fh)
        write_zip(fs, self._temp_path)

    def load_archive(self):
        return zipfs.ZipFS(self._temp_path)

    def remove_archive(self):
        os.remove(self._temp_path)


class TestReadZipFSMem(TestReadZipFS):

    def make_source_fs(self):
        return open_fs('mem://')


class TestDirsZipFS(unittest.TestCase):

    def test_implied(self):
        """Test zipfs creates intermediate directories."""
        fh, path = tempfile.mkstemp('testzip.zip')
        try:
            os.close(fh)
            _zip_file = zipfile.ZipFile(path, mode='w')
            _zip_file.writestr('foo/bar/baz/egg', b'hello')
            _zip_file.close()
            zip_fs = zipfs.ZipFS(path)
            zip_fs.getinfo('foo')
            zip_fs.getinfo('foo/bar')
            zip_fs.getinfo('foo/bar/baz')
            self.assertTrue(zip_fs.isdir('foo/bar/baz'))
            self.assertTrue(zip_fs.isfile('foo/bar/baz/egg'))
        finally:
            os.remove(path)






from __future__ import unicode_literals

import datetime
import unittest

import pytz

from fs.enums import ResourceType
from fs.info import Info
from fs.permissions import Permissions
from fs.time import datetime_to_epoch


class TestInfo(unittest.TestCase):

    def test_empty(self):
        """Test missing info."""
        info = Info({
            'basic': {},
            'details': {},
            'access': {},
            'link': {}
        })

        self.assertIsNone(info.name)
        self.assertIsNone(info.is_dir)
        self.assertEqual(info.type, ResourceType.unknown)
        self.assertIsNone(info.accessed)
        self.assertIsNone(info.modified)
        self.assertIsNone(info.created)
        self.assertIsNone(info.metadata_changed)
        self.assertIsNone(info.accessed)
        self.assertIsNone(info.permissions)
        self.assertIsNone(info.user)
        self.assertIsNone(info.group)
        self.assertIsNone(info.target)
        self.assertFalse(info.is_link)

    def test_access(self):
        info = Info({
            "access": {
                "uid": 10,
                "gid": 12,
                "user": 'will',
                "group": 'devs',
                "permissions": ['u_r']
            }
        })
        self.assertIsInstance(info.permissions, Permissions)
        self.assertEqual(info.permissions, Permissions(user='r'))
        self.assertEqual(info.user, 'will')
        self.assertEqual(info.group, 'devs')
        self.assertEqual(info.uid, 10)
        self.assertEqual(info.gid, 12)

    def test_link(self):
        info = Info({
            'link': {
                'target': 'foo'
            }
        })
        self.assertTrue(info.is_link)
        self.assertEqual(info.target, 'foo')

    def test_basic(self):
        # Check simple file
        info = Info({
            "basic": {
                "name": "bar",
                "is_dir": False
            }
        })
        self.assertEqual(info.name, "bar")
        self.assertIsInstance(info.is_dir, bool)
        self.assertFalse(info.is_dir)
        self.assertEqual(repr(info), "<file 'bar'>")

        # Check dir
        info = Info({
            "basic": {
                "name": "foo",
                "is_dir": True
            }
        })
        self.assertTrue(info.is_dir)
        self.assertEqual(repr(info), "<dir 'foo'>")

    def test_details(self):
        dates = [
            datetime.datetime(2016, 7, 5, tzinfo=pytz.UTC),
            datetime.datetime(2016, 7, 6, tzinfo=pytz.UTC),
            datetime.datetime(2016, 7, 7, tzinfo=pytz.UTC),
            datetime.datetime(2016, 7, 8, tzinfo=pytz.UTC)
        ]
        epochs = [datetime_to_epoch(d) for d in dates]

        info = Info({
            "details": {
                "accessed": epochs[0],
                "modified": epochs[1],
                "created": epochs[2],
                "metadata_changed": epochs[3],
                "type": int(ResourceType.file)
            }
        })
        self.assertEqual(info.accessed, dates[0])
        self.assertEqual(info.modified, dates[1])
        self.assertEqual(info.created, dates[2])
        self.assertEqual(info.metadata_changed, dates[3])
        self.assertIsInstance(info.type, ResourceType)
        self.assertEqual(info.type, ResourceType.file)
        self.assertEqual(info.type, 2)

    def test_has_namespace(self):
        info = Info({
            "basic": {},
            "details": {}
        })
        self.assertTrue(info.has_namespace('basic'))
        self.assertTrue(info.has_namespace('details'))
        self.assertFalse(info.has_namespace('access'))

    def test_copy(self):
        info = Info({
            "basic": {
                "name": "bar",
                "is_dir": False
            }
        })
        info_copy = info.copy()
        self.assertEqual(info.raw, info_copy.raw)

    def test_get(self):
        info = Info({'baz':{}})
        self.assertIsNone(info.get('foo', 'bar'))
        self.assertIsNone(info.get('baz', 'bar'))



# coding: utf-8
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import socket
import ftplib
import os
import platform
import shutil
import tempfile
import time
import unittest
import uuid

from six import text_type

from ftplib import error_perm
from ftplib import error_temp

from pyftpdlib.authorizers import DummyAuthorizer

from fs import errors
from fs.opener import open_fs
from fs.ftpfs import FTPFS, ftp_errors
from fs.test import FSTestCases


# Prevent socket timeouts from slowing tests too much
socket.setdefaulttimeout(1)


class TestFTPFSClass(unittest.TestCase):

    def test_parse_ftp_time(self):
        self.assertIsNone(FTPFS._parse_ftp_time('notreallyatime'))
        t = FTPFS._parse_ftp_time('19740705000000')
        self.assertEqual(t, 142214400)

    def test_parse_mlsx(self):
        info = list(
            FTPFS._parse_mlsx(['create=19740705000000;modify=19740705000000; /foo'])
        )[0]
        self.assertEqual(info['details']['modified'], 142214400)
        self.assertEqual(info['details']['created'], 142214400)

        info = list(FTPFS._parse_mlsx(['foo=bar; ..']))
        self.assertEqual(info, [])

    def test_opener(self):
        ftp_fs = open_fs('ftp://will:wfc@ftp.example.org')
        self.assertIsInstance(ftp_fs, FTPFS)
        self.assertEqual(ftp_fs.host, 'ftp.example.org')


class TestFTPErrors(unittest.TestCase):
    """Test the ftp_errors context manager."""

    def test_manager(self):
        mem_fs = open_fs('mem://')

        with self.assertRaises(errors.ResourceError):
            with ftp_errors(mem_fs, path='foo'):
                raise error_temp

        with self.assertRaises(errors.OperationFailed):
            with ftp_errors(mem_fs):
                raise error_temp

        with self.assertRaises(errors.InsufficientStorage):
            with ftp_errors(mem_fs):
                raise error_perm('552 foo')

        with self.assertRaises(errors.ResourceNotFound):
            with ftp_errors(mem_fs):
                raise error_perm('501 foo')

        with self.assertRaises(errors.PermissionDenied):
            with ftp_errors(mem_fs):
                raise error_perm('999 foo')


class TestFTPFS(FSTestCases, unittest.TestCase):

    user = 'user'
    pasw = '1234'

    @classmethod
    def setUpClass(cls):
        from pyftpdlib.test import ThreadedTestFTPd
        super(TestFTPFS, cls).setUpClass()

        cls._temp_dir = tempfile.mkdtemp('ftpfs2tests')
        cls._temp_path = os.path.join(cls._temp_dir, text_type(uuid.uuid4()))
        os.mkdir(cls._temp_path)

        cls.server = ThreadedTestFTPd()
        cls.server.shutdown_after = -1
        cls.server.handler.authorizer = DummyAuthorizer()
        cls.server.handler.authorizer.add_user(
            cls.user, cls.pasw, cls._temp_path, perm="elradfmw")
        cls.server.handler.authorizer.add_anonymous(cls._temp_path)
        cls.server.start()

        # Don't know why this is necessary on Windows
        if platform.system() == 'Windows':
            time.sleep(0.1)
        # Poll until a connection can be made
        if not cls.server.is_alive():
            raise RuntimeError("could not start FTP server.")

    @classmethod
    def tearDownClass(cls):
        cls.server.stop()
        shutil.rmtree(cls._temp_dir)
        super(TestFTPFS, cls).tearDownClass()

    def make_fs(self):
        return open_fs('ftp://{}:{}@{}:{}'.format(
            self.user, self.pasw, self.server.host, self.server.port
        ))

    def tearDown(self):
        shutil.rmtree(self._temp_path)
        os.mkdir(self._temp_path)
        super(TestFTPFS, self).tearDown()

    def test_ftp_url(self):
        self.assertTrue(self.fs.ftp_url.startswith('ftp://127.0.0.1'))

    def test_host(self):
        self.assertEqual(self.fs.host, self.server.host)

    #@attr('slow')
    def test_connection_error(self):
        fs = FTPFS('ftp.not.a.chance', timeout=1)
        with self.assertRaises(errors.RemoteConnectionError):
            fs.listdir('/')

        with self.assertRaises(errors.RemoteConnectionError):
            fs.makedir('foo')

        with self.assertRaises(errors.RemoteConnectionError):
            fs.open('foo.txt')

    def test_getmeta_unicode_path(self):
        self.assertTrue(self.fs.getmeta().get('unicode_paths'))
        self.fs.features
        del self.fs.features['UTF8']
        self.assertFalse(self.fs.getmeta().get('unicode_paths'))


class TestFTPFSNoMLSD(TestFTPFS):

    def make_fs(self):
        ftp_fs = super(TestFTPFSNoMLSD, self).make_fs()
        ftp_fs.features
        del ftp_fs.features['MLST']
        return ftp_fs

    def test_features(self):
        pass



# coding: utf-8
"""Defines the MemOpener."""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from .base import Opener
from .errors import OpenerError
from ..subfs import ClosingSubFS
from .. import appfs


class AppFSOpener(Opener):

    protocols = [
        'userdata',
        'userconf',
        'sitedata',
        'siteconf',
        'usercache',
        'userlog'
    ]

    _protocol_mapping = {
        'userdata': appfs.UserDataFS,
        'userconf': appfs.UserConfigFS,
        'sitedata': appfs.SiteDataFS,
        'siteconf': appfs.SiteConfigFS,
        'usercache': appfs.UserCacheFS,
        'userlog': appfs.UserLogFS
    }

    def open_fs(self, fs_url, parse_result, writeable, create, cwd):

        fs_class = self._protocol_mapping[parse_result.protocol]
        resource, delim, path = parse_result.resource.partition('/')
        tokens = resource.split(':', 3)
        if len(tokens) == 2:
            appname, author = tokens
            version = None
        elif len(tokens) == 3:
            appname, author, version = tokens
        else:
            raise OpenerError(
                'resource should be <appname>:<author> '
                'or <appname>:<author>:<version>'
            )

        app_fs = fs_class(
            appname,
            author=author,
            version=version,
            create=create
        )

        app_fs = (
            app_fs.opendir(path, factory=ClosingSubFS)
            if delim
            else app_fs
        )

        return app_fs




"""
An abstract permissions container.

"""

from __future__ import print_function
from __future__ import unicode_literals

import six


def make_mode(init):
    """
    Make a mode integer from an initial value.

    """
    return Permissions.get_mode(init)


class _PermProperty(object):
    """Creates simple properties to get/set permissions."""
    def __init__(self, name):
        self._name = name
        self.__doc__ = "Boolean for '{}' permission.".format(name)

    def __get__(self, obj, obj_type=None):
        return self._name in obj

    def __set__(self, obj, value):
        if value:
            obj.add(self._name)
        else:
            obj.remove(self._name)


@six.python_2_unicode_compatible
class Permissions(object):
    """
    An abstraction for file system permissions.

    :param list names: A list of permissions.
    :param int mode: A mode integer.
    :param str user: A triplet of *user* permissions, e.g. ``"rwx"`` or
        ``"r--"``
    :param str group: A triplet of *group* permissions, e.g. ``"rwx"``
        or ``"r--"``
    :param str other: A triplet of *other* permissions, e.g. ``"rwx"``
        or ``"r--"``
    :param bool sticky: A boolean for the *sticky* bit.
    :param bool setuid: A boolean for the *setuid* bit.
    :param bool setguid: A boolean for the *setuid* bit.

    Permissions objects store information regarding the permissions
    on a resource. It supports Linux permissions, but is generic enough
    to manage permission information from almost any filesystem.

    >>> from fs.permissions import Permissions
    >>> p = Permissions(user='rwx', group='rw-', other='r--')
    >>> print(p)
    rwxrw-r--
    >>> p.mode
    500
    >>> oct(p.mode)
    '0764'



    """

    _LINUX_PERMS = [
        ('setuid', 2048),
        ('setguid', 1024),
        ('sticky', 512),
        ('u_r', 256),
        ('u_w', 128),
        ('u_x', 64),
        ('g_r', 32),
        ('g_w', 16),
        ('g_x', 8),
        ('o_r', 4),
        ('o_w', 2),
        ('o_x', 1)
    ]
    _LINUX_PERMS_NAMES = [_name for _name, _mask in _LINUX_PERMS]

    def __init__(self,
                 names=None,
                 mode=None,
                 user=None,
                 group=None,
                 other=None,
                 sticky=None,
                 setuid=None,
                 setguid=None):
        if names is not None:
            self._perms = set(names)
        elif mode is not None:
            self._perms = {
                name
                for name, mask in self._LINUX_PERMS
                if mode & mask
            }
        else:
            perms = self._perms = set()
            perms.update('u_' + p for p in user or '' if p != '-')
            perms.update('g_' + p for p in group or '' if p != '-')
            perms.update('o_' + p for p in other or '' if p != '-')

        if sticky:
            self._perms.add('sticky')
        if setuid:
            self._perms.add('setuid')
        if setguid:
            self._perms.add('setguid')

    def __repr__(self):
        if not self._perms.issubset(self._LINUX_PERMS_NAMES):
            _perms_str = ", ".join(
                "'{}'".format(p) for p in sorted(self._perms)
            )
            return "Permissions(names=[{}])".format(_perms_str)

        def _check(perm, name):
            return name if perm in self._perms else ''

        user = ''.join((
            _check('u_r', 'r'),
            _check('u_w', 'w'),
            _check('u_x', 'x')
        ))
        group = ''.join((
            _check('g_r', 'r'),
            _check('g_w', 'w'),
            _check('g_x', 'x')
        ))
        other = ''.join((
            _check('o_r', 'r'),
            _check('o_w', 'w'),
            _check('o_x', 'x')
        ))
        args = []
        _fmt = "user='{}', group='{}', other='{}'"
        basic = _fmt.format(user, group, other)
        args.append(basic)
        if self.sticky:
            args.append('sticky=True')
        if self.setuid:
            args.append('setuid=True')
        if self.setuid:
            args.append('setguid=True')
        return "Permissions({})".format(', '.join(args))

    def __str__(self):
        return self.as_str()

    def __iter__(self):
        return iter(self._perms)

    def __contains__(self, permission):
        return permission in self._perms

    def __eq__(self, other):
        if isinstance(other, Permissions):
            names = other.dump()
        else:
            names = other
        return self.dump() == names

    def __ne__(self, other):
        return not self.__eq__(other)

    @classmethod
    def parse(cls, ls):
        """Parse permissions in linux notation."""
        user = ls[:3]
        group = ls[3:6]
        other = ls[6:9]
        return cls(user=user, group=group, other=other)

    @classmethod
    def load(cls, permissions):
        """Load a serialized permissions object."""
        return cls(names=permissions)

    @classmethod
    def create(cls, init=None):
        """
        Create a permissions object from an initial value.

        :param init: May be None for equivalent for 0o777 permissions,
            a mode integer, or a list of permission names.
        :returns: mode integer, that may be used by `os.makedir`
            (amongst others).

        >>> Permissions.create(None)
        Permissions(user='rwx', group='rwx', other='rwx')
        >>> Permissions.create(0o700)
        Permissions(user='rwx', group='', other='')
        >>> Permissions.create(['u_r', 'u_w', 'u_x'])
        Permissions(user='rwx', group='', other='')

        """
        if init is None:
            return cls(mode=0o777)
        if isinstance(init, cls):
            return init
        if isinstance(init, int):
            return cls(mode=init)
        if isinstance(init, list):
            return cls(names=init)
        raise ValueError('permissions is invalid')

    @classmethod
    def get_mode(cls, init):
        """
        Convert an initial value to a mode integer.

        """
        return cls.create(init).mode

    def copy(self):
        """Make a copy of this permissions object."""
        return Permissions(names=list(self._perms))

    def dump(self):
        """Get a list suitable for serialization."""
        return sorted(self._perms)

    def as_str(self):
        """Get a linux-style string representation of permissions."""
        perms = [
            c if name in self._perms else '-'
            for name, c in zip(
                self._LINUX_PERMS_NAMES[-9:],
                'rwxrwxrwx'
            )
        ]
        if 'setuid' in self._perms:
            perms[2] = 's' if 'u_x' in self._perms else 'S'
        if 'setguid' in self._perms:
            perms[5] = 's' if 'g_x' in self._perms else 'S'
        if 'sticky' in self._perms:
            perms[8] = 't' if 'o_x' in self._perms else 'T'

        perm_str = ''.join(perms)
        return perm_str

    @property
    def mode(self):
        """Mode integer."""
        mode = 0
        for name, mask in self._LINUX_PERMS:
            if name in self._perms:
                mode |= mask
        return mode

    @mode.setter
    def mode(self, mode):
        """Set mode integer."""
        self._perms = {
            name
            for name, mask in self._LINUX_PERMS
            if mode & mask
        }

    u_r = _PermProperty('u_r')
    u_w = _PermProperty('u_w')
    u_x = _PermProperty('u_x')

    g_r = _PermProperty('g_r')
    g_w = _PermProperty('g_w')
    g_x = _PermProperty('g_x')

    o_r = _PermProperty('o_r')
    o_w = _PermProperty('o_w')
    o_x = _PermProperty('o_x')

    sticky = _PermProperty('sticky')
    setuid = _PermProperty('setuid')
    setguid = _PermProperty('setguid')

    def add(self, *permissions):
        """
        Add permission(s).

        :param permissions: Permission name(s).

        """
        self._perms.update(permissions)

    def remove(self, *permissions):
        """
        Remove permission(s).

        :param permissions: Permission name(s).

        """
        self._perms.difference_update(permissions)

    def check(self, *permissions):
        """
        Check if one or more permissions are enabled.

        :param permissions: Permission name(s).
        :returns: True if all given permissions are set.
        :rtype bool:

        """
        return self._perms.issuperset(permissions)
